"""
Linear Prediction (LP) functions for extrapolating and modeling NMR signals.
"""

from __future__ import print_function

__developer_info__ = """
Notes
^^^^^
This module contains functions for performing linear prediction on NMR  data.
The algorithms used were selected for simplicity to show how linear prediction
works not for computational speed nor stability.  Locations where significant
improvements can be made to improve speed or stability are indicated with
SPEED and STABILITY within the source code with discussion following.

The notation for the Linear Prediction equation, coefficients, roots, etc.
closely match those in "NMR Data Processing" by Hoch snd Stern. This book was
references for many of the algorithms in this module.

Reduced order LP-SVD and LP-TLS methods are not implemented but should
be easy to add if desired.  See :py:func:`find_lproots_hsvd` for an example.

"""

import numpy as np
import scipy.linalg


#############################################
# linear prediction extrapolation functions #
#############################################


def lp(data, pred=1, slice=slice(None), order=8, mode="f", append="after",
       bad_roots="auto", fix_mode="on", mirror=None, method="svd"):
    """
    Linear prediction extrapolation of 1D or 2D data.

    Parameters
    ----------
    data : ndarray
        1D or 2D NMR data with the last (-1) axis in the time domain.
    pred : int
        Number of points to predict along the last axis.
    slice : slice object, optional
        Slice object which selects the region along the last axis to use in LP
        equation.  The default (slice(None)) will use all points.
    order : int
        Prediction order, number of LP coefficients calculated.
    mode : {'f', 'b', 'fb' or 'bf'}
        Mode to generate LP filter. 'f' for forward,'b' for backward, fb for
        'forward-backward and 'bf' for backward-forward.
    append : {'before', 'after'}
        Location to append the data, either 'before' the current data, or
        'after' the existing data. This is independent of the `mode` parameter.
    bad_roots : {'incr', 'decr', None, 'auto'}
        Type of roots which to consider bad and to stabilize.  Option are those
        with increasing signals 'incr' or decreasing signals 'decr'.  None will
        perform no root stabiliting.  The default ('auto') will set the
        parameter based on the `mode` parameter.  'f' or 'fb' `mode` will
        results in a 'incr' `bad_roots` parameter, 'b' or 'bf` in 'decr'
    fix_mode : {'on', 'reflect'}
        Method used to stabilize bad roots, 'on' to move the roots onto the
        unit circle, 'reflect' to reflect bad roots across the unit circle.
        This parameter is ignored when `bad_roots` is None.
    mirror : {None, '0', '180'}
        Mode to form mirror image of data before processing.  None will
        process the data trace as provided (no mirror image). '0' or '180'
        forms a mirror image of the sliced trace to calculate the LP filter.
        '0' should be used with data with no delay, '180' with data
        with an initial half-point delay.
    method : {'svd', 'qr', 'choleskey', 'tls'}
        Method to use to calculate the LP filter. Choices are a SVD ('svd'), QR
        ('qr'), or Choleskey ('choleskey') decomposition, or Total Least
        Squares ('tls').

    Returns
    -------
    ndata : ndarray
        NMR data with `pred` number of points linear predicted and appended to
        the original data.

    Notes
    -----
    When given 2D data a series of 1D linear predictions are made to
    each row in the array, extending each by pred points. To perform a 2D
    linear prediction using a 2D prediction matrix use :py:func:`lp2d`.

    In forward-backward or backward-forward mode root stabilizing is done
    on both sets of signal roots as calculated in the first mode direction.
    After averaging the coefficient the roots are again stabilized.

    When the append parameter does not match the LP mode, for example
    if a backward linear prediction (mode='b') is used to predict points
    after the trace (append='after'), any root fixing is done before reversing
    the filter.

    """
    if data.ndim == 1:
        return lp_1d(data, pred, slice, order, mode, append, bad_roots,
                     fix_mode, mirror, method)
    elif data.ndim == 2:
        # create empty array to hold output
        s = list(data.shape)
        s[-1] = s[-1] + pred
        new = np.empty(s, dtype=data.dtype)
        # vector-wise 1D LP
        for i, trace in enumerate(data):
            new[i] = lp_1d(trace, pred, slice, order, mode, append, bad_roots,
                           fix_mode, mirror, method)
        return new
    else:
        raise ValueError("data must be a one or two dimensional array")


# method specific extrapolation
def lp_svd(data, pred=1, slice=slice(None), order=8, mode="f", append="after",
           bad_roots="auto", fix_mode="on", mirror=None):
    """
    Linear Prediction extrapolation of 1D or 2D data using SVD decomposition.

    See :py:func:`lp` for documentation.

    """
    return lp(data, pred, slice, order, mode, append, bad_roots, fix_mode,
              mirror, method="svd")


def lp_qr(data, pred=1, slice=slice(None), order=8, mode="f", append="after",
          bad_roots="auto", fix_mode="on", mirror=None):
    """
    Linear Prediction extrapolation of 1D or 2D data using QR decomposition.

    See :py:func:`lp` for documentation

    """
    return lp(data, pred, slice, order, mode, append, bad_roots, fix_mode,
              mirror, method="qr")


def lp_cho(data, pred=1, slice=slice(None), order=8, mode="f", append="after",
           bad_roots="auto", fix_mode="on", mirror=None):
    """
    Linear Prediction extrapolation of 1D or 2D data using Cholesky
    decomposition.

    See :py:func:`lp` for documentation
    """
    return lp(data, pred, slice, order, mode, append, bad_roots, fix_mode,
              mirror, method="cholesky")


def lp_tls(data, pred=1, slice=slice(None), order=8, mode="f", append="after",
           bad_roots="auto", fix_mode="on", mirror=None):
    """
    Linear Prediction extrapolation of 1D or 2D data using Total Least Squares.

    See :py:func:`lp` for documentation.

    """
    return lp(data, pred, slice, order, mode, append, bad_roots, fix_mode,
              mirror, method="tls")


# underlying 1D extrapolation
def lp_1d(trace, pred=1, slice=slice(None), order=8, mode="f", append="after",
          bad_roots="auto", fix_mode="on", mirror=None, method="svd"):
    """
    Linear Prediction extrapolation of 1D data.

    Parameters
    ----------
    trace : ndarray
        1D NMR data in the time domain.
    pred : int
        Number of points to predict along the last axis.
    slice : slice object, optional
        Slice object which selects the region along the last axis to use in LP
        equation.  The default (slice(None)) will use all points.
    order : int
        Prediction order, number of LP coefficients calculated.
    mode : {'f', 'b', 'fb' or 'bf'}
        Mode to generate LP filter. 'f' for forward,'b' for backward, fb for
        'forward-backward and 'bf' for backward-forward.
    append : {'before', 'after'}
        Location to append the data, either 'before' the current data, or
        'after' the existing data. This is independent of the `mode` parameter.
    bad_roots : {'incr', 'decr', None, 'auto'}
        Type of roots which to consider bad and to stabilize.  Option are those
        with increasing signals 'incr' or decreasing signals 'decr'.  None will
        perform no root stabiliting.  The default ('auto') will set the
        parameter based on the `mode` parameter.  'f' or 'fb' `mode` will
        results in a 'incr' `bad_roots` parameter, 'b' or 'bf` in 'decr'
    fix_mode : {'on', 'reflect'}
        Method used to stabilize bad roots, 'on' to move the roots onto the
        unit circle, 'reflect' to reflect bad roots across the unit circle.
        This parameter is ignored when `bad_roots` is None.
    mirror : {None, '0', '180'}
        Mode to form mirror image of data before processing.  None will
        process the data trace as provided (no mirror image). '0' or '180'
        forms a mirror image of the sliced trace to calculate the LP filter.
        '0' should be used with data with no delay, '180' with data
        with an initial half-point delay.
    method : {'svd', 'qr', 'choleskey', 'tls'}
        Method to use to calculate the LP filter. Choices are a SVD ('svd'), QR
        ('qr'), or Choleskey ('choleskey') decomposition, or Total Least
        Squares ('tls').

    Returns
    -------
    ntrace : ndarray
        NMR data with `pred` number of points linear predicted and appended to
        the original data.

    Notes
    -----
    In forward-backward or backward-forward mode root stabilizing is done
    on both sets of signal roots as calculated in the first mode direction.
    After averaging the coefficient the roots are again stabilized.

    When the append parameter does not match the LP mode, for example
    if a backward linear prediction (mode='b') is used to predict points
    after the trace (append='after'), any root fixing is done before reversing
    the filter.


    See Also
    --------
    lp : 1D or 2D linear prediction extrapolation.

    """
    # check for bad arguments
    if mode not in ["f", "b", "fb", "bf"]:
        raise ValueError("mode must be 'f', 'b', 'fb', or 'bf'")
    if append not in ["before", "after"]:
        raise ValueError("append must be 'before' or 'after'")
    if bad_roots not in [None, "incr", "decr", "auto"]:
        raise ValueError("bad_roots must be None, 'auto', 'incr' or 'decr'")
    if fix_mode not in ["on", "reflect"]:
        raise ValueError("fix_mode must be 'on' or 'reflect'")
    if mirror not in [None, "0", "180"]:
        raise ValueError("mirror must be None, '0' or '180'")
    if method not in ['svd', 'qr', 'cholesky', 'tls']:
        raise ValueError("Invalid method")
    if trace.ndim != 1:
        raise ValueError("trace must be 1D")

    # bad_roots auto mode
    if bad_roots == "auto":
        if mode == "f" or mode == "fb":
            bad_roots = "incr"
        else:
            bad_roots = "decr"

    x = trace[slice]    # extract region to use for finding LP coefficients

    if mirror is not None:  # make mirror image if selected
        x = make_mirror(x, mirror)

    if mode == "fb":
        a = find_lpc_fb(x, order, bad_roots, fix_mode, method)
        mode = "f"
    elif mode == "bf":
        a = find_lpc_bf(x, order, bad_roots, fix_mode, method)
        mode = "b"
    else:
        # form the LP equation matrix and vector
        D, d = make_Dd(x, order, mode)
        a = find_lpc(D, d, method)  # determind the LP prediction filter

    # stablize roots if needed
    if bad_roots is not None:           # stablize roots if needed
        poles = find_roots(a, mode)  # find roots (poles)
        poles = fix_roots(poles, bad_roots, fix_mode)  # fix roots
        # reverse filter when calculated filter is in wrong direction
        if (mode == "b" and append == "after") or (mode == "f" and
                                                   append == "before"):
            poles = [1. / pole for pole in poles]
            mode = {'f': 'b', 'b': 'f'}[mode]
        a = find_coeff(poles, mode)  # find LP filter from roots
    else:
        # reverse filter when calculated filter is in wrong direction
        if (mode == "b" and append == "after") or (mode == "f" and
                                                   append == "before"):
            a = reverse_filter(a, mode)
    # extrapolate the trace using the prediction filter
    ntrace = extrapolate(trace, a, pred, append)
    return ntrace


##################
# LP2D functions #
##################


def lp2d(data, pred, P, M, mirror='0', fix_points=True, method='svd'):
    """
    Perform a forward 2D linear prediction extrapolation on data.

    Use the 2D linear prediction algorithm presented in:
    G. Zhu and A. Bax, Journal of Magnetic Resonance, 1992, 98, 192-199.
    to extend the last (1) axis by `pred` points. A PxM prediction matrix, C,
    is formed by solving the modified linear prediction equation given by:

    data[n,m] = /sigma_{l=0}^{P-1} /sigma_{k=1}^M C_{l,k}*data[n-l,m-k]

    For all valid points in data. This prediction matrix together with the
    data matrix with a mirror image appended is used to extend the last (1)
    axis by pred points resulting in a new array of size [N_0, N_1+pred] where
    N_0 and N_1 are the sizes of the original data.  To linear predict
    both dimensions this function should be used twice with a transpose
    between the uses.

    Backward linear prediction using this method is not possible as the
    method depends on being able to mirror the data before the first collected
    point. In backwards mode this would correspond to being able to correctly
    determind points after the last point which cannot be determinded using the
    mirror method. A backward prediction matrix can be calculated but would
    not prove useful.

    The forward-backward averaging of the linear prediction coefficients is
    not possible as there is no characteristic polynomial to root and reflect.
    Therefore the backward prediction matrix cannot be reversed.

    Parameters
    ----------
    data : ndarray
        2D NMR data (time domain for last axes).
    pred : int
        Number of points to predict along the last (1) axes.
    P : int
        Prediction matrix length along the non-predicted (0) axis.
    M : int
        Prediction matrix length along the predicted (1) axis.
    mirror : {'0' or '180'}
        Method to use for forming the mirror image of the non-predicted axis.
        '0' indicated no initial delay, '180' for a half-point delay.
    fix_points : bool
        True to reduce predicted points with magnitude larger than the largest
        data point. False leaved predicted points  unaltered.
    method : {'svd', 'qr', 'cholesky', 'tls'}
        Method used to calculate the LP prediction matrix.  See :py:func:`lp`
        for a description of theses methods.

    Returns
    -------
    ndata : ndarray
        2D NMR data with `pred` points appended to the last (1) axes.

    Notes
    -----
    The axes in this function are reversed as compared to the JMR paper.

    """
    # check parameters
    if data.ndim != 2:
        raise ValueError("data must be a 2D array")
    if mirror not in ['0', '180']:
        raise ValueError("mirror must be '0' or '180'")
    if method not in ['svd', 'qr', 'cholesky', 'tls']:
        raise ValueError("method must be 'svd', 'qr', 'cholesky' or 'tls'")

    # form lp2d equation matrix and vector
    D, d = make_lp2d_Dd(data, P, M, 'f')

    # Solve lp2d equation to find the prediction matrix
    c = find_lpc(D, d, method)
    C = c.reshape(P, M)

    # extrapolate the 2D data using the prediction matrix
    return extrapolate_2d(data, C, pred, fix_points, mirror)


def extrapolate_2d(x, C, pred, fix_points, mirror):
    """
    Extrapolate points along the 1st axis using the lp2d algorithm.
    """
    # find the prediction matrix shape and flatten it
    P, M = C.shape
    c = C.flatten()

    # find data parameters
    x_max = x.max()
    N_0, N_1 = x.shape

    # create a empty matrix
    if mirror == "0":
        new = np.empty((2 * N_0 - 1, N_1 + pred), dtype=x.dtype)
        plane = N_0 - 1   # index of first non-mirrored point
    else:
        new = np.empty((2 * N_0, N_1 + pred), dtype=x.dtype)
        plane = N_0     # index of first non-mirrored plane

    last = new.shape[0]     # number of rows in new matrix

    # fill the matrix with the mirrored version of each column
    for i in range(N_1):
        new[:, i] = make_mirror(x[:, i], mirror)

    # fill each new column with predicted values
    # i,j give coordinates of top-left corner of PxM reading matrix
    # after filling a column, replace the whole column with the mirrored column

    for j in range(N_1 - M, N_1 - M + pred):  # column index loop
        for i in range(plane - P + 1, last - P + 1):  # row index loop
            new[i + P - 1, j + M] = np.sum(new[i:i + P, j:j + M].flat * c)

            if fix_points:  # reduce predicted point is needed
                if new[i + P - 1, j + M] > x_max:
                    new[i + P - 1, j + M] = ((x_max * x_max) /
                                             (new[i + P - 1, j + M]))

        # fill the column with the mirrored column so it can be read in the
        # next interation of the loop
        new[:, j + M] = make_mirror(new[plane:, j + M], mirror)

    return new[plane:]


def make_lp2d_Dd(x, P, M, mode='f'):
    """
    Form the lp2d equation matrix and vector.
    """
    # Form the D and d' matrix and vector in the 2DLP equation:
    # D*c = d'
    # where c is a flattened prediction matrix ordered:
    # (P-1,M) (P-1,M-1) ... (P-1,1) (P-2,M-1) ... (P-2,1) (P-3,M-1) ... ...
    # (2,1) (1,M-1) ... (1,1) (0,M-1) ... (0,1)
    if mode == 'b':
        # backward mode not implemented
        # this would have the d' value as the
        # top left corner
        # this can be done with the same code
        # after reversing x,
        # x = x[::-1,::-1]
        raise NotImplemented

    # Build D and d' row by row by flattening a PxM region of the x matrix
    # starting at 0, 0 and moving down to the bottom of the matrix, then moving
    # back to the top and over one row, and continue till all data has been
    # read. d' is filled with the element next to the right corner of this
    # PxM region.

    N_0, N_1 = x.shape  # length of the matrix

    count_P = N_0 - P + 1   # number of valid starting position vertically
    # number of valid starting position horizontally
    # taking into account the element next to the
    # bottom right corner is the predicted value.
    count_M = N_1 - M

    # create an empty D matrix
    D = np.empty((count_P * count_M, P * M), dtype=x.dtype)
    d = np.empty((count_P * count_M, 1), dtype=x.dtype)

    # fill D and d' row by row (i,j) give coordinates of top-left corner of
    # PxM reading matrix
    for j in range(count_M):
        for i in range(count_P):
            D[j * count_P + i] = x[i:i + P, j:j + M].flat
            d[j * count_P + i] = x[i + P - 1, j + M]
    return D, d


#############################################
# Cadzow/Minumum variance signal enhacement #
#############################################


def cadzow(data, M, K, niter, min_var=False):
    """
    Perform a (row wise) Cadzow-like signal enhancement on 1D or 2D data.

    Performs a Cadzow-like signal enhancement with optional adjustment
    of singular values using the minimum variance method as desribed in:
    Chen, VanHuffel, Decanniere, VanHecke, JMR, 1994, 109A, 46-55.

    For 2D data performs independant enhancement on each row of data array.

    Parameters
    ----------
    data : ndarray
        1D or 2D NMR data to enhance.
    M : int
        Large prediction order.  For best results should be between
        K + 5 and 2 * K.
    K : int
        Reduced prediction order.
    niter : int
        Number if iteration of the Cadzow procedure to perform.
    min_var : bool
        True to adjust retained singular values using the  minimum variance
        method. False does not correct the singular values and is the
        Cadzow method.

    Returns
    -------
    ndata : ndarray
        Array of enhanced data

    """
    if data.ndim == 1:
        for i in range(niter):
            data = cadzow_single(data, M, K, min_var)
        return data
    elif data.ndim == 2:
        for trace in data:
            for i in range(niter):
                trace = cadzow_single(trace, M, K, min_var)
        return data
    else:
        raise ValueError("data must be a 1D or 2D array")


def cadzow_single(x, M, K, min_var=False):
    """
    Perform a single iteration of Cadzow signal enhancement on a 1D vector

    See :py:func:`cadzow` for documentation.

    """
    # variable names based upon Chen et al, JMR 1994 109A 46
    # form the Hankel data matrix X
    N = len(x)
    L = N - M + 1
    X = scipy.linalg.hankel(x[:L], x[L - 1:])

    # Compute the SVD of X
    U, s, Vh = scipy.linalg.svd(X)

    # correct the singular values and truncate the rank K
    Ul = np.mat(U[:, :K])
    Vlh = np.mat(Vh[:K, :])     # first K columns of V are first K rows of Vh
    sl = s[:K]

    if min_var:  # adjust singular values using minimum variance method
        # estimate variance of noise singular values
        s2 = (1. / (M - K)) * np.power(s[K:], 2).sum()
        sl = np.array([l - s2 / l for l in sl])

    Sl = np.mat(np.diag(sl))

    # compute enhanced data vector for rank-reduced data matrix
    Xp = Ul * Sl * Vlh

    xp = np.empty_like(x)
    for i, v in enumerate(range(M - 1, -L, -1)):
        # the anti-diagonal is the diagonal with rows reversed
        xp[i] = np.diag(Xp[:, ::-1], v).mean()
    return xp


###################################################
# Linear Prediction parametric modeling functions #
###################################################


def lp_model(trace, slice=slice(None), order=8, mode="f", mirror=None,
             method="svd", full=False):
    """
    Use Linear Prediction to model 1D NMR time domain data.

    Parameters
    ----------
    trace : 1D ndarray
        One dimensional time domain NMR data to model.
    slice : slice object, optional
        Slice object which selects the region along the last axis to use in LP
        equation.  The default, slice(None), will use all points.
    order : int
        Prediction order, number of LP coefficients calculated.
    mode : {'f', 'b'}
        Mode to generate LP filter. 'f' for forward,'b' for backward.
    mirror : {None, '0', '180'}
        Mode to form mirror image of data before processing.  None will
        process the data trace as provided (no mirror image). '0' or '180'
        forms a mirror image of the sliced trace to calculate the LP filter.
        '0' should be used with data with no delay, '180' with data
        with an initial half-point delay.
    method : {'svd', 'qr', 'choleskey', 'tls'}
        Method to use to calculate the LP filter. Choices are a SVD ('svd'), QR
        ('qr'), or Choleskey ('choleskey') decomposition, or Hankel SVD
        ('hsvd').
    full : bool
        True to return amplitudes and phases calculated by performing a least
        squares fitting to the data after LP modeling.  False will return only
        the damping (relaxation) factors and signal frequencies.

    Returns
    -------
    damp : list
        List of damping (relaxation) factors found from LP modeling.
    freq : list
        List of signal frequencies found from LP modeling.
    amp : list, optional
        List of signal amplitudes found by least squares fitting of data after
        LP modeling, only returned when `full` parameter is True.
    phase : list, optional.
        List of signal phases found by least squares fitting of data after LP
        modeling, only returned when `full` parameter is True.


    Notes
    -----
    When backward LP is used the signal roots are reflected before calculating
    model parameters.

    """
    # check for bad arguments
    if mode not in ["f", "b"]:
        raise ValueError("mode must be 'f' or 'b'")
    if method not in ['svd', 'qr', 'cholesky', 'tls', 'hsvd']:
        raise ValueError("Invalid method")
    if trace.ndim != 1:
        raise ValueError("trace must be a 1D array")

    x = trace[slice]    # extract region to use for finding LP coefficients

    if mirror is not None:  # make mirror image if requested
        x = make_mirror(x, mirror)

    # calculate LP coefficient and factor to find poles
    if method in ['svd', 'qr', 'cholseky', 'tls']:
        D, d = make_Dd(x, order, mode)   # form the LP equation elements
        a = find_lpc(D, d, method)      # find LP coefficients
        poles = find_roots(a, mode)     # find roots

    elif method == "hsvd":
        poles = find_lproots_hsvd(x, M=order, K=order, mode=mode, zmethod='sm')
    else:
        raise ValueError("Invalid method")

    # reverse poles if we have backward poles
    if mode == "b":
        poles = [1. / pole for pole in poles]

    # determind the damping factor and frequencies from the roots
    damp = [root2damp(pole) for pole in poles]
    freq = [root2freq(pole) for pole in poles]

    if full is False:
        return damp, freq

    # perform Least Squares fitting to determind amplitudes and phases.

    # We need to find a least squares solutions to:
    # z_0*b_0^0+z_1*b_1^0+.... = x_0
    # z_0*b_0^1+z_1*b_1^1+.... = x_1
    # ...
    # Where b is the LP roots (poles), x is the signal, and a_0 are unknown
    #
    # To solve this recast into Bz = x and solve for a using scipy.lstsq

    # SPEED
    # B is a Vandermonde matrix, this characteristic can be used to more
    # efficiently solve this least squares problem.

    # build the B matrix (a Vandermonde matrix) and solve for the coefficients
    poles = np.array(poles)
    B = np.row_stack([poles ** (i) for i in range(len(x))])
    z, resid, rank, s = np.linalg.lstsq(B, np.array(x))

    # Now the z_n = amp_n*exp(phase_n*i), use this to determind the amplitudes
    # and phases
    amp = [cof2amp(cof) for cof in z]
    phase = [cof2phase(cof) for cof in z]

    return damp, freq, amp, phase


###############################################################
# functions to determine signal parameters from LP parameters #
###############################################################


def root2damp(pole):
    """
    Calculate the damping factor from a LP root
    """
    # damping factor is the radius in the complex plane -1/pi*ln(|pole|)
    return -1. / (np.pi) * np.log(np.abs(pole))


def root2freq(pole):
    """
    Calculate the frequency from a LP root
    """
    # frequency is the angle from the x-axis to the point in the complex plane
    # arg(pole) / (2*pi) = atan2(imag,real) / (2*pi)
    return np.arctan2(pole.imag, pole.real) / (2. * np.pi)


def cof2amp(z):
    """
    Calculate a signal amplitude from a model coefficient
    """
    # z = amp*exp(phase*i) so amp is abs(z)
    return np.abs(z)


def cof2phase(z):
    """
    Calculate a signal phase from a model coefficient
    """
    # z = amp*exp(phase(i) so phase is the arg(z)
    return np.arctan2(z.imag, z.real)


##############################
# data preperation functions #
##############################


def make_D(x, order, mode):
    """
    Make the LP equation D matrix (Da = d')
    """
    L = len(x) - order
    if mode == "f":
        return scipy.linalg.hankel(x[:L], x[L - 1:-1])
    elif mode == "b":
        return scipy.linalg.hankel(x[1:L + 1], x[L:])
    else:
        raise ValueError("mode must be 'f' or 'b'")


def make_little_d(x, order, mode):
    """
    Make the LP equation d' vector (Da = d')
    """
    if mode == "f":
        return x[order:].reshape(len(x) - order, 1)
    elif mode == "b":
        L = len(x) - order
        return x[:L].reshape(L, 1)
    else:
        raise ValueError("mode must be 'f' or 'b'")


def make_Dd(x, order, mode):
    """
    make the LP equation D matrix and d' vector (Da=d')
    """
    return make_D(x, order, mode), make_little_d(x, order, mode)


def make_mirror(x, mode):
    """
    Make a mirror image trace.

    Reflects trace over zero as described in:
    G. Zhu and A. Bax, Journal of Magnetic Resonance, 1990, 90, 405

    When mode is "0" (no initial delay) form the an array with length 2N-1:
        x_n-1 ... x_1 x_0 x_1 ... x_n-1

    When mode is "180" (half point delay) form an array with length 2N:
        x_n-1 .. x_1 x_0 x_0 x_1 ... x_n-1


    Parameters
    ----------
    x : ndarray
        1D array to form mirrored trace from.
    mode : {'180', '0'}
        Mirror mode, see above.

    """
    if mode == "0":
        return np.concatenate((x[:0:-1], x))
    elif mode == "180":
        return np.concatenate((x[::-1], x))


###########################################################
# LP prediction filter calculation functions (find_lpc_*) #
###########################################################

# the coefficients returned from these functions depend on the mode of the
# prediction.  Forward LP returns coefficients ordered m, m-1, ...1
# Backward LP returns 1, 2, ..., m where m is the order of the prediction.


def find_lpc(D, d, method):
    """
    Find linear prediction filter using a provided method.
    """
    if method == "svd":
        return find_lpc_svd(D, d)
    elif method == "qr":
        return find_lpc_qr(D, d)
    elif method == "cholesky":
        return find_lpc_cholesky(D, d)
    elif method == "tls":
        return find_lpc_tls(D, d)
    else:
        raise ValueError("invalid method")


def find_lpc_svd(D, d):
    """
    Find linear prediction filter using single value decomposition.
    """
    L = D.shape[0]
    m = D.shape[1]
    U, s, Vh = scipy.linalg.svd(D)  # SVD decomposition
    U, Vh = np.mat(U), np.mat(Vh)   # make U and Vh matrices
    Si = pinv_diagsvd(s, m, L)      # construct the pseudo-inverse sigma matrix
    return np.array(Vh.H * Si * U.H * d)

# the next 3 lines and the pinv_diagsvd function were adapted from the
# scipy.linalg.pinv2 function - jjh
eps = np.finfo('float').eps
feps = np.finfo('single').eps
_array_precision = {'f': 0, 'd': 1, 'F': 0, 'D': 1}


def pinv_diagsvd(s, m, L):
    """
    Construct the pseudo-inverse of the sigma matrix from singular values
    """
    t = s.dtype.char
    cond = {0: feps * 1e3, 1: eps * 1e6}[_array_precision[t]]
    cutoff = s[0] * cond
    Si = np.zeros((m, L), t)
    for i in range(len(s)):
        if s[i] > cutoff:
            Si[i, i] = 1.0 / np.conj(s[i])
    return Si


def find_lpc_qr(D, d):
    """
    Find linear prediction filter using QR decomposition.
    """
    q, r = scipy.linalg.qr(D)
    q, r = np.mat(q), np.mat(r)

    # SPEED
    # the next line is slow and the use of pinv2 should be avoided as
    # pseudo inversion of r involves a computationally expensive SVD
    # decomposition which is not needed.  Rather r*x = q.H*d should be
    # solved for x using LAPACK's ZTRTRS function (or similar function with
    # different prefix).  This is not currently available in scipy/numpy and
    # therefore is not used here.
    return scipy.linalg.pinv2(r) * q.H * d


def find_lpc_cholesky(D, d):
    """
    Find linear prediction filter using a Cholesky decomposition.
    """
    # form the normal equation (D.H*D)*a = D.H*d
    # SPEED
    # this can be improved by using the Hankel nature of D
    D = np.mat(D)
    DhD = np.mat(np.dot(D.H, D))
    Dhd = np.mat(np.dot(D.H, d))

    c, lower = scipy.linalg.cho_factor(DhD)     # Compute Cholesky decomp.
    return scipy.linalg.cho_solve((c, lower), Dhd)  # solve normal equation


def find_lpc_tls(D, d):
    """
    Find linear prediction filter using the Total Least Squares method
    """
    m = D.shape[1]  # the order of the prediction
    E = np.append(D, d, axis=1)     # form the augmented data matrix
    U, s, Vh = scipy.linalg.svd(E)  # SVD decompositon of augmented matrix
    V = np.conj(Vh.T)               # Hermetian transpose
    return (-1. / V[m, m] * V[:m, m]).reshape((m, 1))


def find_lpc_fb(x, order, bad_roots, fix_mode, method):
    """
    Determind LP coefficient using forward-backward linear prediction.

    Averages LP coefficients generated from solving the forward and backward
    linear prediction equations after reversing the roots of characteristic
    polynomial of the backward solution.  Method is described in:
    G. Zhu and A. Bax, Journal of Magnetic Resonance, 1992, 100, 202-207.

    Description of parameters can be found in :py:func:`lp`.

    """
    # find forward LP coefficients
    D, d = make_Dd(x, order, 'f')
    a = find_lpc(D, d, method)

    # stabilize roots if needed
    if bad_roots is not None:
        poles = find_roots(a, 'f')
        poles = fix_roots(poles, bad_roots, fix_mode)
        a = find_coeff(poles, 'f')
    # store the forward coefficients
    forward_a = a.copy()

    # find the backwards LP coefficients
    D, d = make_Dd(x, order, 'b')
    a = find_lpc(D, d, method)

    # find poles, reverse poles
    poles = find_roots(a, 'b')
    poles = [1. / pole for pole in poles]
    # stabilize roots if needed
    if bad_roots is not None:
        poles = fix_roots(poles, bad_roots, fix_mode)
    # find the backward predicted, forward ordered coefficients
    backward_a = find_coeff(poles, 'f')

    # average the forward and backward coefficients
    return (forward_a + backward_a) / 2.


def find_lpc_bf(x, order, bad_roots, fix_mode, method):
    """
    Determind LP coefficient using backward-forward linear prediction.

    Averages LP coefficients generated from solving the forward and backward
    linear prediction equations after reversing the roots of characteristic
    polynomial of the forward solution.  Similar to method described in:
    G. Zhu and A. Bax, Journal of Magnetic Resonance, 1992, 100, 202-207.

    Description of parameters can be found in :py:func:`lp` function.

    """
    # find backward LP coefficients
    D, d = make_Dd(x, order, 'b')
    a = find_lpc(D, d, method)

    # stabilize roots if needed
    if bad_roots is not None:
        poles = find_roots(a, 'b')
        poles = fix_roots(poles, bad_roots, fix_mode)
        a = find_coeff(poles, 'b')
    # store the forward coefficients
    backward_a = a.copy()

    # find the forward LP coefficients
    D, d = make_Dd(x, order, 'f')
    a = find_lpc(D, d, method)

    # find poles, reverse poles
    poles = find_roots(a, 'f')
    poles = [1. / pole for pole in poles]
    # stabilize roots if needed
    if bad_roots is not None:
        poles = fix_roots(poles, bad_roots, fix_mode)
    # find the forward predicted, backward ordered coefficients
    forward_a = find_coeff(poles, 'b')

    # average the forward and backward coefficients
    return (forward_a + backward_a) / 2.


#####################################
# root finding and fixing functions #
#####################################


def find_lproots_hsvd(x, M, K, mode, zmethod='sm'):
    """
    Find LP roots (poles) using the HSVD method

    Perform a HSVD linear prediction to determind signal roots (poles) as
    described in:
    Barkhuijsen, DeBeer, and Van Ormondt, JMR, 1987, 73, 553

    Parameters x, M and K are the same as those described in the above article.
    zmethod refer to the method used to calculate Z', either a least-squares
    method (lstsq) can be used to solve U_b*Z'=U_t or the Sherman-Morrison
    formula (sm) can be used to avoid the full matrix inversion with equation
    [12] being used to find Z'. The Sherman-Morrison method should be faster
    with similar precision.

    Parameters
    ----------
    x : 1D ndarray
        1D trace of NMR data in the time domain, the FID.
    M : int
        Length (M+1) of data matrix to form.
    K : int
        Reduced prediction order (number of signal roots) Must be less than the
        smaller of M + 1 or len(x) - M.
    mode : {'f', 'b'}
        Mode to perform LP. 'f' for forward,'b' for backward.
    zmethod : {'lstsq', 'sm'}
        Method used to find Z' 'lstsq' for least squares, 'sm' for
        Sherman-Morrison.

    Returns
    -------
    y : ndarray
        Array of signal roots (poles)

    """
    # check parameters
    if mode not in ['f', 'b']:
        raise ValueError("mode must be 'f' or 'b'")
    if zmethod not in ['lstsq', 'sm']:
        raise ValueError("zmethod must be 'lstsq' or 'sm'")
    if K > min(M + 1, len(x) - M):
        raise ValueError("K must be less than min(M + 1, len(x) - M)")

    # form the data matrix X
    N = len(x)
    L = N - M - 1

    if mode == "f":
        X = scipy.linalg.hankel(x[:L + 1], x[L:])
    else:
        # for backward LP we need to make the hankel matrix:
        # x_N-1 x_N-2 ... x_N-M-1
        # x_N-2 x_N-3 ... x_N-M-2
        # ...
        # x_M   x_M-1 ... x_0
        X = scipy.linalg.hankel(x[:M - 1:-1], x[M::-1])

    # SVD of data matrix and truncation of U to form Uk
    U, s, Vh = scipy.linalg.svd(X)
    Uk = np.mat(U[:, :K])   # trucated U matrix of rank K
    Ub = Uk[:-1]            # Uk with bottom row removed
    Ut = Uk[1:]             # Uk with top row removed

    # calculate the Z' matrix
    if zmethod == 'lstsq':  # solve Ub*Z' = Ut using least-squares
        Zp, resid, rank, s = scipy.linalg.lstsq(Ub, Ut)
    else:
        # solve using equation [12]:
        # Z' = (Ek + (u*uh / (1-uh*u)) ) * Ub.H*Ut
        uh = Uk[-1]  # bottom row of Uk
        u = uh.H
        Zp = (np.eye(K, dtype=u.dtype) + (u * uh / (1. - uh * u))) * Ub.H * Ut

    # diagonalization (find eigenvalues) of Z' to yield roots
    return scipy.linalg.eigvals(Zp)


def find_roots(a, mode="f"):
    """
    Find LP roots (poles) from a set of LP coefficients.

    Parameters
    ----------
    a : array
        LP coefficients.
    mode : {'f', 'b'}
        Mode of LP coefficients.  'f' for coefficients ordered m, m - 1,..., 1.
        'b' for coefficients ordered 1, 2, ...., m

    Returns
    -------
    roots : array
        LP roots (poles)

    """
    if mode not in ['f', 'b']:
        raise ValueError("mode must be 'f' or 'b'")

    # STABILITY
    # the algorithm here is that used by numpy roots, build the companion
    # matrix and find its eigenvalues.  These values should be polished for
    # better numerical accuracy.

    # np.roots expects a array, p, with coefficients
    # p[0] * x**n + p[1] * x**(n-1] + ... + p[n-1]*x + p[n]
    # in forward mode LP the coefficients are ordered m,m-1,...1
    # in backward mode LP the coefficient are ordered is 1,2,...,m
    # To calculate the roots, create a leading 1.0+0.0j and reverse if needed.
    p = np.empty(len(a) + 1, dtype=a.dtype)
    p[0] = (1.0 + 0.0j)

    if mode == "f":      # reverse for forward LP
        p[1:] = -a.flat[::-1]
    else:   # backward LP
        p[1:] = -a.flat[:]
    return np.roots(p)


def find_coeff(poles, mode="f"):
    """
    Find LP coefficients from a set of LP roots (poles).

    Parameters
    ----------
    poles : ndarray
        Array of LP roots (poles)
    mode : {'f', 'b'}
        Mode in which LP coefficients should be returned.  'f' for coefficients
        ordered m, m - 1,..., 1. 'b' for coefficients ordered 1, 2, ...., m.

    Returns
    -------
    c : ndarray
        LP coefficients ordered according to `mode`.

    """
    # STABILITY
    # the algorithm used here is numpy poly function which convolves
    # the roots to find the coefficients, the accuracy of this method
    # depends on the dtype of the poles parameter.
    if mode not in ['f', 'b']:
        raise ValueError("mode must be 'f'or 'b'")

    if mode == 'f':  # reverse resulting coefficients
        return np.squeeze(-np.poly(poles)[:0:-1])
    else:   # keep coefficients as is
        return np.squeeze(-np.poly(poles)[1:])


def reverse_filter(a, mode):
    """
    Reverse a filter (change forward LP to backwards LP).
    """
    nmode = {'f': 'b', 'b': 'f'}[mode]
    # find roots, replace each root with 1/root, then recalculate filter
    return find_coeff([1. / pole for pole in find_roots(a, mode)], nmode)


def fix_roots(poles, fix_roots="incr", fix_mode="reflect"):
    """
    Fix (stabilize) LP roots.

    Parameters
    ----------
    poles : ndarray
        Array of LP roots (poles).
    fix_roots : {'incr', 'decr'}
        Type of roots which to consider bad and to stabilize.  Either those
        with increasing signals 'incr' or decreasing signals 'decr'.
    fix_mode : {'on', 'reflect'}
        Method used to stabilize bad roots, 'on' to move the roots onto the
        unit circle, 'reflect' to reflect bad roots across the unit circle.

    Returns
    -------
    npoles : ndarray
        Array of stabilized LP roots (poles).

    """
    if fix_roots not in ["incr", "decr"]:
        raise ValueError("fix_roots must be 'incr' or 'decr'")
    if fix_mode not in ["on", "reflect"]:
        raise ValueError("fix_mode must be 'on' or 'reflect'")

    if fix_roots == "incr":     # remove increasing signals
        for i, pole in enumerate(poles):
            if np.abs(pole) > 1:
                # print("Fixing root:",i)
                if fix_mode == "on":
                    poles[i] = pole / np.abs(pole)
                else:
                    poles[i] = 1 / np.conj(pole)
    else:   # remove decreasing signals
        for i, pole in enumerate(poles):
            if np.abs(pole) < 1:
                # print("Fixing root:",i)
                if fix_mode == "on":
                    poles[i] = pole / np.abs(pole)
                else:
                    poles[i] = 1 / np.conj(pole)
    return poles


###########################
# Extrapolation functions #
###########################


def extrapolate(trace, a, pred, append):
    """
    Extrapolate points using LP prediction filter.

    Parameters
    ----------
    trace : 1D ndarray
        1D array to extrapolate from and append to.
    a : ndarray
        LP coefficients, must be ordered according to direction of
        extrapolation.
    pred : int
        Number of points to predict using LP.
    append : {'a', 'b'}
        Location to append new points, 'a' for after the current data, 'b' for
        before the current data.

    Returns
    -------
    ntrace : 1D ndarray
        1D array with extrapolated points appended

    """
    m = len(a)      # LP order
    M = len(trace)   # number of points in original trace
    ntrace = np.empty((M + pred), dtype=trace.dtype)

    if append not in ["after", "before"]:
        raise ValueError("append must be 'a' or 'b'")

    if append == "after":   # append after trace
        ntrace[:M] = trace
        for i in range(pred):
            ntrace[M + i] = np.sum(np.multiply(ntrace[M - m + i:M + i],
                                   a.flat))
        return ntrace

    if append == "before":   # append before trace
        ntrace[-M:] = trace
        for i in range(pred):
            ntrace[pred - i - 1] = np.sum(np.multiply(
                ntrace[pred - i:pred + m - i], a.flat))
        return ntrace



"""
Functions for reading Tecmag .tnt data files.
"""
__developer_doc__ = """
Tecmag .tnt file format information
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
The Tecmag .tnt file format is documented with C pseudo-code in the file
"A1 - TNMR File Format.doc" distributed with the TNMR software.

This file is based on the
`pytnt module <https://www.github.com/chatcannon/pytnt>`_.
Please inform upstream if you find a bug or to request additional features.

"""

import io
import os
import re

import numpy as np

from . import fileiobase


TNTMAGIC_RE = re.compile(b"^TNT1\.\d\d\d$")

TNTMAGIC = np.dtype('a8')
TNTTLV = np.dtype([('tag', 'a4'), ('bool', '<u4'), ('length', '<u4')])

TNTTMAG = np.dtype([
    ('npts', '<i4', 4),
    ('actual_npts', '<i4', 4),
    ('acq_points', '<i4'),
    ('npts_start', '<i4', 4),
    ('scans', '<i4'),
    ('actual_scans', '<i4'),
    ('dummy_scans', '<i4'),
    ('repeat_times', '<i4'),
    ('sadimension', '<i4'),
    ('samode', '<i4'),
    # ('space1', 'a0'),

    ('magnet_field', '<f8'),
    ('ob_freq', '<f8', 4),
    ('base_freq', '<f8', 4),
    ('offset_freq', '<f8', 4),
    ('ref_freq', '<f8'),
    ('NMR_frequency', '<f8'),
    ('obs_channel', '<i2'),
    ('space2', 'a42'),

    ('sw', '<f8', 4),
    ('dwell', '<f8', 4),
    ('filter', '<f8'),
    ('experiment_time', '<f8'),
    ('acq_time', '<f8'),
    ('last_delay', '<f8'),
    ('spectrum_direction', '<i2'),
    ('hardware_sideband', '<i2'),
    ('Taps', '<i2'),
    ('Type', '<i2'),
    ('bDigRec', '<u4'),
    ('nDigitalCenter', '<i4'),
    ('space3', 'a16'),

    ('transmitter_gain', '<i2'),
    ('receiver_gain', '<i2'),
    ('NumberOfReceivers', '<i2'),
    ('RG2', '<i2'),
    ('receiver_phase', '<f8'),
    ('space4', 'a4'),

    ('set_spin_rate', '<u2'),
    ('actual_spin_rate', '<u2'),

    ('lock_field', '<i2'),
    ('lock_power', '<i2'),
    ('lock_gain', '<i2'),
    ('lock_phase', '<i2'),
    ('lock_freq_mhz', '<f8'),
    ('lock_ppm', '<f8'),
    ('H2O_freq_ref', '<f8'),
    ('space5', 'a16'),

    ('set_temperature', '<f8'),
    ('actual_temperature', '<f8'),

    ('shim_units', '<f8'),
    ('shims', '<i2', 36),
    ('shim_FWHM', '<f8'),

    ('HH_dcpl_attn', '<i2'),
    ('DF_DN', '<i2'),
    ('F1_tran_mode', '<i2', 7),
    ('dec_BW', '<i2'),
    ('grd_orientation', 'a4'),
    ('LatchLP', '<i4'),
    ('grd_Theta', '<f8'),
    ('grd_Phi', '<f8'),
    ('space6', 'a264'),

    ('start_time', '<u4'),
    ('finish_time', '<u4'),
    ('elapsed_time', '<i4'),

    ('date', 'a32'),
    ('nuclei', 'a16', 4),
    ('sequence', 'a32'),
    ('lock_solvent', 'a16'),
    ('lock_nucleus', 'a16')
])


TNTGRIDANDAXIS = np.dtype([
    ('majorTickInc', '<f8', 12),
    ('minorIntNum', '<i2', 12),
    ('labelPrecision', '<i2', 12),
    ('gaussPerCentimeter', '<f8'),
    ('gridLines', '<i2'),
    ('axisUnits', '<i2'),
    ('showGrid', '<u4'),
    ('showGridLabels', '<u4'),
    ('adjustOnZoom', '<u4'),
    ('showDistanceUnits', '<u4'),
    ('axisName', 'a32'),
    ('space', 'a52'),
])


TNTTMG2 = np.dtype([
    ('real_flag', '<u4'),
    ('imag_flag', '<u4'),
    ('magn_flag', '<u4'),
    ('axis_visible', '<u4'),
    ('auto_scale', '<u4'),
    ('line_display', '<u4'),
    ('show_shim_units', '<u4'),

    ('integral_display', '<u4'),
    ('fit_display', '<u4'),
    ('show_pivot', '<u4'),
    ('label_peaks', '<u4'),
    ('keep_manual_peaks', '<u4'),
    ('label_peaks_in_units', '<u4'),
    ('integral_dc_average', '<u4'),
    ('integral_show_multiplier', '<u4'),
    ('Boolean_space', '<u4', 9),

    ('all_ffts_done', '<u4', 4),
    ('all_phase_done', '<u4', 4),

    ('amp', '<f8'),
    ('ampbits', '<f8'),
    ('ampCtl', '<f8'),
    ('offset', '<i4'),

    ('axis_set', TNTGRIDANDAXIS),

    ('display_units', '<i2', 4),
    ('ref_point', '<i4', 4),
    ('ref_value', '<f8', 4),
    ('z_start', '<i4'),
    ('z_end', '<i4'),
    ('z_select_start', '<i4'),
    ('z_select_end', '<i4'),
    ('last_zoom_start', '<i4'),
    ('last_zoom_end', '<i4'),
    ('index_2D', '<i4'),
    ('index_3D', '<i4'),
    ('index_4D', '<i4'),

    ('apodization_done', '<i4', 4),
    ('linebrd', '<f8', 4),
    ('gaussbrd', '<f8', 4),
    ('dmbrd', '<f8', 4),
    ('sine_bell_shift', '<f8', 4),
    ('sine_bell_width', '<f8', 4),
    ('sine_bell_skew', '<f8', 4),
    ('Trapz_point_1', '<i4', 4),
    ('Trapz_point_2', '<i4', 4),
    ('Trapz_point_3', '<i4', 4),
    ('Trapz_point_4', '<i4', 4),
    ('trafbrd', '<f8', 4),
    ('echo_center', '<i4', 4),

    ('data_shift_points', '<i4'),
    ('fft_flag', '<i2', 4),
    ('unused', '<f8', 8),
    ('pivot_point', '<i4', 4),
    ('cumm_0_phase', '<f8', 4),
    ('cumm_1_phase', '<f8', 4),
    ('manual_0_phase', '<f8'),
    ('manual_1_phase', '<f8'),
    ('phase_0_value', '<f8'),
    ('phase_1_value', '<f8'),
    ('session_phase_0', '<f8'),
    ('session_phase_1', '<f8'),

    ('max_index', '<i4'),
    ('min_index', '<i4'),
    ('peak_threshold', '<f4'),
    ('peak_noise', '<f4'),
    ('integral_dc_points', '<i2'),
    ('integral_label_type', '<i2'),
    ('integral_scale_factor', '<f4'),
    ('auto_integrate_shoulder', '<i4'),
    ('auto_integrate_noise', '<f8'),
    ('auto_integrate_threshold', '<f8'),
    ('s_n_peak', '<i4'),
    ('s_n_noise_start', '<i4'),
    ('s_n_noise_end', '<i4'),
    ('s_n_calculated', '<f4'),

    ('Spline_point', '<i4', 14),
    ('Spline_point_avr', '<i2'),
    ('Poly_point', '<i4', 8),
    ('Poly_point_avr', '<i2'),
    ('Poly_order', '<i2'),

    ('space', 'a610'),

    ('line_simulation_name', 'a32'),
    ('integral_template_name', 'a32'),
    ('baseline_template_name', 'a32'),
    ('layout_name', 'a32'),
    ('relax_information_name', 'a32'),
    ('username', 'a32'),
    ('user_string_1', 'a16'),
    ('user_string_2', 'a16'),
    ('user_string_3', 'a16'),
    ('user_string_4', 'a16')
])


def read(filename):
    """
    Read a Tecmag .tnt data file.

    Parameters
    ----------
    filename : str
        Name of file to read from

    Returns
    -------
    dic : dict
        Dictionary of Tecmag parameters.
    data : ndarray
        Array of NMR data.

    """
    tnt_sections = dict()

    with open(filename, 'rb') as tntfile:

        tntmagic = np.fromstring(tntfile.read(TNTMAGIC.itemsize),
                                 TNTMAGIC, count=1)[0]

        if not TNTMAGIC_RE.match(tntmagic):
            err = ("Invalid magic number (is '%s' really TNMR file?): %s" %
                   (filename, tntmagic))
            raise ValueError(err)

        # Read in the section headers
        tnthdrbytes = tntfile.read(TNTTLV.itemsize)
        while(TNTTLV.itemsize == len(tnthdrbytes)):
            tlv = np.fromstring(tnthdrbytes, TNTTLV)[0]
            data_length = tlv['length']
            hdrdict = {'offset': tntfile.tell(),
                       'length': data_length,
                       'bool': bool(tlv['bool'])}
            if data_length <= 4096:
                hdrdict['data'] = tntfile.read(data_length)
                assert(len(hdrdict['data']) == data_length)
            else:
                tntfile.seek(data_length, os.SEEK_CUR)
            tnt_sections[tlv['tag'].decode()] = hdrdict
            tnthdrbytes = tntfile.read(TNTTLV.itemsize)

    assert(tnt_sections['TMAG']['length'] == TNTTMAG.itemsize)
    tmag = np.fromstring(tnt_sections['TMAG']['data'], TNTTMAG, count=1)[0]

    assert(tnt_sections['DATA']['length'] ==
           tmag['actual_npts'].prod() * 8)
    #  For some reason we can't set offset and shape together
    # DATA = np.memmap(tntfilename,np.dtype('<c8'), mode='r',
    #                  offset=self.tnt_sections['DATA']['offset'],
    #                  shape=self.TMAG['actual_npts'].tolist(),order='F')
    data = np.memmap(filename, np.dtype('<c8'), mode='c',
                     offset=tnt_sections['DATA']['offset'],
                     shape=tmag['actual_npts'].prod())
    data = np.reshape(data, tmag['actual_npts'], order='F')

    assert(tnt_sections['TMG2']['length'] == TNTTMG2.itemsize)
    tmg2 = np.fromstring(tnt_sections['TMG2']['data'], TNTTMG2, count=1)[0]

    dic = dict()
    for name in TNTTMAG.names:
        if not name.startswith('space'):
            dic[name] = tmag[name]
    for name in TNTTMG2.names:
        if name not in ['Boolean_space', 'unused', 'space', 'axis_set']:
            dic[name] = tmg2[name]
    for name in TNTGRIDANDAXIS.names:
        dic[name] = tmg2['axis_set'][name]

    return dic, data


def guess_udic(dic, data):
    """
    Guess parameters of universal dictionary from dic, data pair.

    Parameters
    ----------
    dic : dict
        Dictionary of Tecmag parameters.
    data : ndarray
        Array of NMR data.

    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.

    """

    # create an empty universal dictionary
    udic = fileiobase.create_blank_udic(4)

    # update default values
    for i in range(4):
        udic[i]["size"] = dic['actual_npts'][i]
        udic[i]["sw"] = dic['sw'][i]
        udic[i]["complex"] = True
        udic[i]["obs"] = dic['ob_freq'] * 1e6
        # Not sure what the difference is here
        # N.B. base_freq is some bogus value like 1.4e-13
        udic[i]["car"] = dic['ob_freq'] * 1e6
        udic[i]["time"] = not bool(dic['fft_flag'][i])
        udic[i]["freq"] = bool(dic['fft_flag'][i])

    return udic



"""
Functions for reading `nmrML <http://nmrml.org/>`_ files.
"""

import base64
import zlib
from warnings import warn

import numpy as np


def read(filename, data_dtype=None):
    """
    Read a nmrML file.

    Parameters
    ----------
    filename : str
        Name of nmrML file to read.
    data_dtype : str, optional
        NumPy data type of the data. None, the default, will determine this
        data type from the infomation in the file. Occasionally this
        information is incorrect and this argument can be used to explicitly
        supply this information.

    Returns
    -------
    dic : dict
        Dictionary of spectra parameters.
    data : ndarray, 1D
        Array of spectral data.

    """
    import xmltodict  # delay import so that xmltodict is optional
    with open(filename, 'rb') as nmrml_file:
        doc = xmltodict.parse(nmrml_file.read())

    fid_dict = doc['nmrML']['acquisition']['acquisition1D']['fidData']
    data = _get_nmrml_data(fid_dict, data_dtype)

    dic = doc['nmrML']
    return dic, data


BYTE_FORMAT_TO_DTYPE = {
    'float32': '<f4',
    'float64': '<f8',
    'complex64': '<c8',
    'complex128': '<c16',
}


def _get_nmrml_data(fid_dict, data_dtype):
    """ Return a NumPy array for spectral data in a nmrML data file. """
    # read required parameters from dictionary
    byte_format = fid_dict['@byteFormat'].lower()
    compressed = fid_dict['@compressed'].lower() == 'true'
    encoded_length = int(fid_dict['@encodedLength'])
    base64_text = fid_dict['#text']
    if len(base64_text) != encoded_length:
        warn("Size of encoded text (%i) does not match stated length (%i)"
             % (len(base64_text), encoded_length))

    # decode and uncompress
    compressed_bytes = base64.b64decode(base64_text)
    if compressed:
        uncompressed_bytes = zlib.decompress(compressed_bytes)
    else:
        uncompressed_bytes = compressed_bytes

    # convert to ndarray
    if data_dtype is None:
        data_dtype = BYTE_FORMAT_TO_DTYPE[byte_format]
    data = np.fromstring(uncompressed_bytes, dtype=data_dtype)

    return data



"""
Test Program

"""
import sys


def main():

    python = sys.version.split("\n")[0]

    try:
        name, curdir, curexpno, curprocno = sys.argv
        message = """
        Welcome to Python {}.
        The current directory is set to {}
        EXPNO {} is currently open at PROCNO {}
        """.format(python, curdir, curexpno, curprocno)

    except ValueError:
        message = """
        Welcome to Python {}.
        No directory is currently open, 
        or none was passed on to this script.
        """.format(python)

    print(message)


if __name__ == "__main__":
    main()



""" Unit tests for nmrglue/analysis/integration.py module """

import numpy as np
from scipy.stats import multivariate_normal
from numpy.testing import assert_array_almost_equal

import nmrglue as ng
from nmrglue.analysis.integration import integrate, ndintegrate


# test helper functions
def _build_1d_ppm_scale():
    n = 500
    sw = 1/1000e-6
    obs = 100  # Mhz

    dx = (sw / n)/obs
    ppm_scale = np.arange(dx, dx * n, dx)

    # To match the apparent NMR convention the array has to be reversed.
    return ppm_scale[::-1]


# test functions
def test_1d_integrate():
    """ Test integration of synthetic 1D data with two peaks."""
    # generate test scale
    ppm_scale = _build_1d_ppm_scale()
    uc = ng.fileio.fileiobase.uc_from_freqscale(ppm_scale, 100)

    # generate test data
    data = (multivariate_normal.pdf(ppm_scale, mean=5, cov=0.01) +
            multivariate_normal.pdf(ppm_scale, mean=8, cov=0.01) * 2.)

    # import matplotlib.pyplot as plt
    # plt.figure()
    # plt.plot(ppm_scale, data)
    # plt.show()

    # Test with a single integral region
    assert integrate(data, uc, (4, 6)) - 1.0 <= np.finfo(float).eps
    assert integrate(data, uc, (7, 9)) - 2.0 <= np.finfo(float).eps

    # Test with multiple integral regions:
    limits = [(4, 6), (7, 9)]
    values_1 = integrate(data, uc, limits)
    assert_array_almost_equal(values_1,  [1., 2.])

    # Test with multiple integral regions and re-normalization of the values
    # to the value of the range between 7 ppm and 9 ppm
    values_2 = integrate(data, uc, limits, norm_to_range=1)
    assert_array_almost_equal(values_2,  [0.5, 1.])

    # Test with multiple integral regions and re-normalization of the values
    # to the value of the range between 7 ppm and 9 ppm and re-calibration
    # of the the value to 3.
    values_2 = integrate(data, uc, limits, norm_to_range=1, calibrate=3)
    assert_array_almost_equal(values_2,  [0.5 * 3, 3.])


def test_1d_integrate_withnoise():
    """ Test integration of synthetic 1D data with two peaks and noise"""

    # seed random number
    np.random.seed(0)

    # generate test scale
    ppm_scale = _build_1d_ppm_scale()
    uc = ng.fileio.fileiobase.uc_from_freqscale(ppm_scale, 100)

    # generate test data
    data = (multivariate_normal.pdf(ppm_scale, mean=5, cov=0.01) +
            multivariate_normal.pdf(ppm_scale, mean=8, cov=0.01) * 2)
    noise = np.random.randn(*data.shape) * 0.01
    data = data + noise

    # import matplotlib.pyplot as plt
    # plt.figure()
    # plt.plot(ppm_scale, data)
    # plt.show()

    # estimate of how much the noise can throw off the the integral
    limits = (4.0, 6.0)
    n = abs(uc(4.5, 'ppm')-uc(5.5, 'ppm'))
    max_error = np.std(noise) * np.sqrt(n)

    # Test with a single integral region sanity check.
    assert abs(integrate(data, uc, limits) - 1) <= max_error

    # Test renormalization of norms
    resutls = integrate(data, uc, ((4, 6), (7, 9)), noise_limits=(1, 2),
                        norm_to_range=1)

    # Test renormalization of values.
    assert abs(resutls[0, 0] - 0.5) <= max_error


def test_1d_ndintegrate():
    """ Test integration of synthetic with ndintegrate.
    """
    # generate test scale
    ppm_scale = _build_1d_ppm_scale()
    uc = ng.fileio.fileiobase.uc_from_freqscale(ppm_scale, 100)

    # generate test data
    data = (multivariate_normal.pdf(ppm_scale, mean=5, cov=0.01) +
            multivariate_normal.pdf(ppm_scale, mean=8, cov=0.01) * 2.)

    # Test with a single  integral region
    assert abs(ndintegrate(data, uc, (4, 6)) - 1.0) <= 1e-10
    assert abs(ndintegrate(data, uc, [(7, 9), ]) - 2.0) <= 1e-10


def test_2d_ndintegrate():
    """ Test integration of synthetic 2D data with ndintegrate. """
    # generate test scale
    ppm_scale = _build_1d_ppm_scale()
    uc = ng.fileio.fileiobase.uc_from_freqscale(ppm_scale, 100)

    # generate test data
    x, y = np.meshgrid(ppm_scale, ppm_scale)
    pos = np.empty(x.shape + (2,))
    pos[:, :, 0] = x
    pos[:, :, 1] = y
    rv = multivariate_normal([5, 8], [0.01, 0.01])
    data = rv.pdf(pos)

    # import matplotlib.pyplot as plt
    # plt.figure()
    # plt.contour(x, y, data)
    # plt.show()

    limits = ((4, 6), (7, 9))

    assert abs(ndintegrate(data, [uc, uc], limits) - 1.0) <= 1e-10



"""
A collection of NMR processing functions for filtering, smoothing, and
correcting spectral baselines.
"""

import numpy as np
import scipy
import scipy.ndimage
import scipy.signal
from scipy.interpolate import InterpolatedUnivariateSpline
from itertools import groupby
from operator import itemgetter


# Linear (First order) baseline correction
def base(data, nl, nw=0):
    """
    Linear (first-order) baseline correction based on node list.

    Parameters
    ----------
    data : 1D or 2D ndarray
        Array of 1D or 2D NMR data.
    nl : list
        List of baseline nodes.
    nw : float, optional
        Node half-width in points.

    Returns
    -------
    ndata : ndarray
        NMR data with first order baseline correction appied.  For 2D data
        baseline correction is applied for each trace along the last
        dimension.

    """
    if data.ndim == 1:
        data = data - calc_bl_linear(data, nl, nw)
    else:   # for 2D array loop over traces
        for i, vec in enumerate(data):
            data[i] = data[i] - calc_bl_linear(vec, nl, nw)
    return data


def calc_bl_linear(x, nl, nw=0):
    """
    Calculate a baseline using linear approximation between nodes

    Parameters
    ----------
    x : 1D ndarray
        One-dimensional NMR data.
    nl : list
        List of baseline nodes
    nw : float
        Node half-width in points

    Returns
    -------
    baseline : ndarray
        Base calculated using linear approximation between nodes.

    """
    bl = np.zeros_like(x)
    for i in range(len(nl) - 1):
        # minimum and maximum index
        min = nl[i]
        max = nl[i + 1]
        # linspace s1 and s2
        s1 = x[min - nw:min + nw + 1].mean()
        s2 = x[max - nw:max + nw + 1].mean()
        bl[min:max + 1] = np.linspace(s1, s2, max - min + 1)
    return bl


# Constant baseline correction
def cbf(data, last=10, apply=slice(None)):
    """
    Constant baseline correction from percentage of data.

    Parameters
    ----------
    data : 1D or 2D ndarray
        Array of 1D or 2D NMR data.
    last : float, optional
        Percent (0 - 100) of last axis used to calculate the baseline
        correction.
    apply : slice, optional
        Slice describing first-axis region(s) to which the baseline correction
        should be applied.  Parameter is ignored for 1D data.

    Returns
    -------
    ndata : 1D or 2D ndarray
        NMR data with a constant baseline subtracted.

    """
    # calculate the correction
    n = data.shape[-1] * last // 100 + 1
    corr = data[..., -n:].sum(axis=-1) / n

    # apply correction
    if data.ndim == 2:
        data[apply] = data[apply] - np.array([corr]).transpose()[apply]
        return data
    else:
        return data - corr


def cbf_explicit(data, calc=slice(None), apply=slice(None)):
    """
    Constant Baseline correction from an explicit region of the data.

    Parameters
    ----------
    data : 1D or 2D ndarray
        Array of 1D or 2D NMR data.
    calc : slice, optional
        Slice describing region to use for calculating the baseline correction.
    apply : slice, optional
        Slice describing first-axis region(s) to which the baseline correction
        should be applied.  Parameter is ignored for 1D data.

    Returns
    -------
    ndata : 1D or 2D ndarray
        NMR data with a constant baseline subtracted.

    """
    # calculate correction
    n = len(range(data.shape[-1])[calc])
    corr = data[..., calc].sum(axis=-1) / n

    # apply correction
    if data.ndim == 2:
        data[apply] = data[apply] - np.array([corr]).transpose()[apply]
        return data
    else:
        return data - corr


# Median baseline correction
def med(data, mw=24, sf=16, sigma=5.0):
    """
    Median baseline correction.

    Algorith described in: Friedrichs, M.S. JBNMR 1995 5 147-153.

    Parameters
    ----------
    data : 1D or 2D ndarray
        Array of 1D or 2D NMR data.
    mw : float
        Median window size in pts.
    sf : float
        Smooth window size in pts.
    sigma : float
        Standard-deviation of convoluted Gaussian window.

    Returns
    -------
    ndata : 1D or 2D ndarray
        NMR data with the median baseline subtracted.

    """
    if data.ndim == 1:
        data = data - calc_bl_med(data, mw, sf, sigma)
    else:
        for i, vec in enumerate(data):
            data[i] = vec - calc_bl_med(vec, mw, sf, sigma)
    return data


def calc_bl_med(x, mw, sf, sigma):
    """
    Calculate a baseline using median baseline correction.

    Algorithm described in: Friedrichs, M.S. JBNMR 1995 5 147-153

    Parameters
    ----------
    x : 1D ndarray
        One dimensional NMR data
    mw : float
        Median window size in pts.
    sf : float
        Smooth window size in pts.
    sigma : float
        Standard-deviation of convoluted Gaussian window.

    Returns
    -------
    baseline : 1D ndarray
        Baseline calculated using median baseline correction

    """
    # create extrema array (non extrema values are masked out)
    mask = x == scipy.ndimage.median_filter(x, size=3)
    mask[0] = False     # first pt always extrema
    mask[-1] = False    # last pt always extrema
    e = np.ma.masked_array(x, mask)

    # fill in the median vector
    m = scipy.ndimage.median_filter(e, mw + 1, mode="mirror")
    # using the median_filter might give slightly different results than
    # described algorithm but is MUCH faster

    # convolve with a gaussian
    g = scipy.signal.gaussian(sf, sigma)
    g = g / g.sum()
    return scipy.signal.convolve(m, g, mode='same')


# Distribution based classification method for calculation of baseline
# in 1D spectra
def baseline_corrector(data, wd=20):
    """
    Calculate a baseline using a distribution based classification method.

    Algorithm described in: Wang et al. Anal. Chem. 2013, 85, 1231-1239

    Parameters
    ----------
    data : 1D ndarray
        One dimensional NMR data with real values
    wd : float
        Median window size in pts.

    Returns
    -------
    data : 1D ndarray
        Baseline corrected spectrum  calculated using distribution based
        classification

    """
    sd_set = _get_sd(data, wd)
    sigma = _find_noise_sd(sd_set, 0.999)
    sn_vector = _is_signal(sigma, sd_set, 3)
    s_start = _get_signal_start(sn_vector)
    s_end = np.sort(len(data) - 1 - _get_signal_start((sn_vector[::-1])))
    r = _get_temp_baseline(data, s_start, s_end, 7)
    baseline = _smooth(r, 60)
    return data - baseline


def _rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def _get_sd(data, k):
    return np.std(_rolling_window(data, k), -1)


def _find_noise_sd(sd_set, ratio):
    '''Calculate the median m1 from SDset. exclude the elements greater
    than 2m1from SDset and recalculate the median m2. Repeat until
    m2/m1 converge(sd_set)'''
    m1 = np.median(sd_set)
    S = sd_set <= 2.0 * m1
    tmp = S * sd_set
    sd_set = tmp[tmp != 0]
    m2 = np.median(sd_set)
    while m2/m1 < ratio:
        m1 = np.median(sd_set)
        S = sd_set <= 2.0 * m1
        tmp = S * sd_set
        sd_set = tmp[tmp != 0]
    return m2


def _is_signal(sigma, sd_set, w):
    sn_vector = sd_set * 0
    for i in np.arange(len(sn_vector)):
        if sd_set[i] > sigma * 1.1:
            sn_vector[np.maximum(0, i-w):np.minimum(i + w, len(sn_vector))] = 1
    return sn_vector


def _get_signal_start(sn_vector):
    s_start = []
    for k, v in groupby(enumerate(sn_vector), key=itemgetter(1)):
        if k:
            v = list(v)
            s_start.append(v[0][0])
    return np.array(s_start)


def _get_temp_baseline(data, s_start, s_end, w):
    xi = np.arange(len(data))
    x = np.vstack((s_start, s_end))[0]
    y = np.convolve(data[x], np.ones((7,)) / 7, mode='same')
    r = InterpolatedUnivariateSpline(x, y, k=1)
    tmp = r(xi)
    return tmp


def _smooth(r, w):
    return scipy.signal.medfilt(r, (2 * w + 1))


# Solvent Filter
def sol_general(data, filter, w=16, mode='same'):
    """
    Solvent filter with generic filter.

    Algorithm described in: Marion et al. JMR 1989 84 425-430

    Parameters
    ----------
    data : 1D or 2D ndarray
        Array of 1D or 2D NMR data.
    filter : ndarray
        Filter to convolve with data.  Not used in solvent filter functions
        which specific the filter, e.g. sol_boxcar.
    w : int, optional
        Filter length.  Not used here but is used in solent filter functions
        which specificy the filter, e.g. sol_boxcar.
    mode : {'valid', 'same', 'full'}, optional
        Convolution mode, 'same' should be used.

    Returns
    -------
    ndata : 1D or 2D ndarray
        NMR data with solvent filter applied

    """
    A = filter.sum()
    if data.ndim == 2:
        filter = np.atleast_2d(filter)
    return data - scipy.signal.convolve(data, filter, mode=mode) / A


def sol_boxcar(data, w=16, mode='same'):
    """
    Solvent filter with boxcar filter.  See :py:func:`sol_general`.
    """
    filter = scipy.signal.boxcar(w)
    return sol_general(data, filter, w=w, mode=mode)


def sol_sine(data, w=16, mode='same'):
    """
    Solvent filter with sine-bell filter. See :py:func:`sol_general`.
    """
    filter = np.cos(np.pi * np.linspace(-0.5, 0.5, w))
    return sol_general(data, filter, w=w, mode=mode)


def sol_sine2(data, w=16, mode='same'):
    """
    Solvent filter with square sine-bell filter. See :py:func:`sol_general`.
    """
    filter = np.cos(np.pi * np.linspace(-0.5, 0.5, w)) ** 2
    return sol_general(data, filter, w=w, mode=mode)


def sol_gaussian(data, w=16, mode='same'):
    """
    Solvent filter with square gaussian filter. See :py:func:`sol_general`.
    """
    filter = scipy.signal.gaussian(w, w / 2.)
    return sol_general(data, filter, w=w, mode=mode)


# Polynomial Solvent Subtraction
def poly_td(data):
    """
    Polynomial time domain solvent subtraction - Not Implemented
    """
    # From NMRPipe paper(appendix):
    #
    # when used with the argument -time, fits all data points to a polynomial,
    # which is then subtracted from the original data.  It is intended to fit
    # and subtract low-freqency solvent signal in the FID, a procedure that
    # often causes less distortions than time-domain convolution methods.
    # By default, a fourth-order polynomials is used.  For speed successive
    # averages of regions are usually fit, rather than fitting all of the data.
    #
    # Algorithm
    #
    # 1. Calculate average of blocks
    # 2. Fit these averages to polynomial (block parameters)
    # 3. Back out "real" polynomial parameters from these block parameters
    # 4. Subtract off the polynomial from data
    raise NotImplementedError


def poly_fd(data):
    """
    Polynomial frequency domain baseline correction - Not Implemented
    """
    # From NMRPipe paper (appendix):
    #
    # applies a polynomial baseline correction of the order specified by
    # argument -ord via an automated base-line detection method when used
    # with argument -auto.  The defauly is a forth-order polynomial. The
    # automated base-line mode works as follows: a copy of a given vector is
    # divided into a series of adjacent sections, typically eight points wide.
    # The average value of each section is subtracted from all points in that
    # section, to generate a 'centered' vector.  The intensities of the entire
    # centered vector are sorted, and the standard deviation of the noise is
    # estimated under the assumption that a given fraction (typically about
    # 30%) of the smallest intensities belong to the base-line, and that the
    # noise is normally distributed.  This noise estimate is multiplied by a
    # constant, typically about 1.5, to yield a classification threshold.
    # Then, each section in the centered vector is classified as base line only
    # if its standard deviation does not exceed the threshold.  These
    # classifications are used to correct the original vector.
    #
    # Algorithm
    #
    # 1. Divide into 'blocks'
    # 2. Center each block and create centered vector
    # 3. Calculate intensity (abs) of each centered vector
    # 4. Sort intensities, lower 30% belong to baseline
    # 5. Fit base line intensities to Normal distribution, gives estimation
    #    of standard deviation (std) of noise
    # 6. Classification threshold set to 1.5*std
    # 7. Qualify each block in centered vector as baseline only
    #    (its std < thres) or not (std > thres)
    # 8. Fit baseline only points to polynomial and substract off
    raise NotImplementedError



"""
Helper functions for lineshape fitting.
"""


# bound functions
def delta(v, d):
    """
    Limits of v +/- d for array v and d.
    """
    return zip(v - d, v + d)


def limit(vmin, vmax, npeaks):
    """
    Fixed limits, vmin to vmax for all peaks.
    """
    return [(vmin, vmax)] * npeaks


def scale_params(rec, prefix, first, last):
    """
    scale lineshape parameters in rec with name prefixX with X from from
    first to last (inclusive)
    """
    return [rec[prefix + str(z)] for z in range(first, last + 1)]


def no_limits(nparams, npeaks):
    """
    No limits on nparameters for npeaks.
    """
    return [[(None, None)] * npeaks] * nparams


def no_limits_amp(npeaks):
    """
    No amplitude limits for npeaks
    """
    return [(None, None)] * npeaks


# misc
def super_zip(l):
    """
    Zip a list after zipping each element.

    Useful for lineshape parameter and bounds.
    """
    return zip(*[zip(*i) for i in l])


def scale_columns(prefix, first, last):
    """
    Create a list of scale parameter column names with name prefixX with X
    from first to last (inclusive)
    """
    return [prefix + str(i) for i in range(first, last + 1)]



from . import bruker
from . import convert
from . import jcampdx
from . import nmrml
from . import pipe
from . import rnmrtk
from . import simpson
from . import sparky
from . import tecmag
from . import varian
# replicate varian namespace in agilent
from . import varian as agilent
from . import table



import numpy as np


def integrate(data, unit_conv, limits, unit='ppm', noise_limits=None,
              norm_to_range=None, calibrate=1.0):
    """
    Integrate one 1D data array within limits given in units. Data points must
    be equally spaced.

    Functional form of integration is:

    .. math::
        value = \sum_a^b s(x_{i}) dx

    Where:
    s is the signal, a and b are the limits of integration and dx is the width
    of each bin.

    The integration error due to baseline noise is calculated as:

    .. math::
        error = \sigma_{vol} = \sigma \sqrt{n}

    if the noise_limits are set.

    Where:

    .. math::
        n =  \\frac{|b-a|}{dx}+1

    sigma is the standard deviation of the baseline noise. n is the number
    of bins in the integration range.

    Parameters
    ----------
    data: array like
        1d array of intensity
    unit_conv: `fileiobase.unit_conversion` object
        unit_conversion object associated with data
    limits: array like
        With shape (2,) or (P, 2). Array with lower and upper integration
        limit. Or array with P rows of lower and upper integration limits.
    noise_limits: Optional[array like]
        With shape(2, ). Array with lower and upper limits to section of data
        with only noise. A larger range will likely yield a more accurate
        estimate. It is unwise to use the very end of the spectrum for the
        noise range.
    norm_to_range: Optional[int]
        If given, all values are normalized to the value determined within the
        range given by limits[norm_to_range, :]
    calibrate: Optional[float]
        If norm_to_range is given, the values are re-normalized so that the
        integral within limits[norm_to_range, :] equals some number.

    Returns
    -------
    array
        [value, ...] integration values

    if noise_limits is given:

    array
        [[value, error], ...] where error a one sigma estimate of the error
        only from the spectrum noise

    """
    # convert form limits in units to lower and upper index
    limits = np.array(limits)
    if limits.size == 2:
        limits = np.expand_dims(limits, axis=0)
    inds = [(unit_conv(x0, unit), unit_conv(x1, unit)) for (x0, x1) in limits]
    inds = np.array([sorted(ind) for ind in inds])

    # sum part of the integral
    sum_slice = np.array([np.sum(data[slice(*ind)]) for ind in inds])

    # dx part of the integral
    scale = unit_conv.ppm_scale()
    dx = abs(scale[1]-scale[0])

    # put together the integral
    values = sum_slice * dx

    if noise_limits is not None:

        # determine the standard deviation of the noise
        noise_inds = (unit_conv(noise_limits[0], unit),
                      unit_conv(noise_limits[1], unit))
        noise_inds = sorted(noise_inds)

        # the error (from noise) is  std * dx * limits range
        std = np.std(data[slice(*noise_inds)])
        errors = std * np.sqrt(inds[:, 1] - inds[:, 0])

        if norm_to_range is not None:
            # You must normalize the error before you normalize the value!
            errors = (errors / values[norm_to_range]) * calibrate
            values = (values / values[norm_to_range]) * calibrate
        return np.vstack((values, errors)).T

    else:

        if norm_to_range is not None:
            values = (values / values[norm_to_range]) * calibrate
        return values


def ndintegrate(data, unit_conv, limits, unit='ppm', noise_limits=None):
    """
    Integrate one nD data array within limits given in units. Data points must
    be equally spaced. Can only integrate one region per function call.

    The integration error due to baseline noise is calculated as:

    .. math::
        error = \sigma_{vol} = \sigma \sqrt{\prod_i^{d} n_{i}},

    if the noise_limits are set.

    Where:
    sigma is the standard deviation of the baseline noise. n is the number
    of bins in the integration range for each d dimensions.

    See integrate for more information.

    Parameters
    ----------
    data: array like
        1d array of intensity
    unit_convs: [`fileiobase.unit_conversion`, ] list
        list of unit_conversion object associated with each dim of data.
    limits: array like
        With shape (2,) or (d, 2). 1D Array with lower and upper integration
        limits for 1D . Or array with d rows of lower and upper integration
        limits for each dimension.
    noise_limits: Optional[array like]
        With shape(2, ). Array with lower and upper limits to section of data
        with only noise. A larger range will likely yield a more accurate
        estimate. It is unwise to use the very end of the spectrum for the
        noise range.

    Returns
    -------
    array
        [value, ...] integration values

    if noise_limits is given:

    array
        [[value, error], ...] where error a one sigma estimate of the error
        only from the spectrum noise
    """

    # determine the dimensionality of the data.
    d = np.ndim(data)

    try:
        iter(unit_conv)
    except TypeError:
        unit_conv = [unit_conv, ]

    if d != len(unit_conv):
        mesg = 'A unit_conversion object is needed for each dimension.'
        raise ValueError(mesg)

    limits = np.array(limits)
    if limits.ndim == 1:
        limits = np.expand_dims(limits, axis=0)

    if limits.shape[0] != d and limits.shape[1] != 2:
        mesg = 'A lower and upper limit is needed for each dimension.'
        raise ValueError(mesg)

    inds = [(uc(x[0], unit), uc(x[1], unit))
            for (uc, x) in zip(unit_conv, limits)]
    inds = [sorted(x) for x in inds]

    # the integrate_nd needs to be scaled by the bin width in ppm
    ppm_scales = [x.ppm_scale() for x in unit_conv]
    dx = np.prod(np.array([abs(x[1]-x[0]) for x in ppm_scales]))

    slice_sum = (data[[slice(x[0], x[1])for x in np.flipud(inds)]]).sum()

    value = slice_sum * dx

    if noise_limits is None:
        return value

    else:
        noise_limits = np.array(noise_limits)
        if noise_limits.size == 2:
            noise_limits = np.expand_dims(noise_limits, axis=0)

        if noise_limits.shape[0] != d and noise_limits.shape[1] != 2:
            mesg = 'If given, a noise limit is needed for each dimension.'
            raise ValueError(mesg)

        noise_inds = [(uc(x[0], unit), uc(x[1], unit))
                      for (uc, x) in zip(unit_conv, noise_limits)]
        noise_inds = [sorted(x) for x in noise_inds]

        # see docstring of integrate
        nm = np.prod(np.array([abs(x[1]-x[0]) for x in noise_inds]))
        std = np.std(data[[slice(x[0], x[1])for x in np.flipud(noise_inds)]])

        error = std * np.sqrt(nm)
        return np.hstack((value, error))



"""
Fuctions for reading and writing Rowland NMR Toolkit (RNMRTK) files
"""

from __future__ import division

__developer_info__ = """
Information of the Rowland NMR Toolkit file format
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

"""

import numpy as np
from warnings import warn

from . import fileiobase

###################
# unit conversion #
###################


def make_uc(dic, data, dim=-1):
    """
    Creat a unit conversion object

    Parameters
    ----------
    dic : dict
        Dictionary of RNMRTK parameters.
    data : ndarray
        Array of NMR data.
    dim : int, optional
        Demension number to create unit conversion object for. Default is for
        the last dimension.

    Returns
    -------
    uc : unit conversion object.
        Unit conversion object for given dimension.

    """
    if dim < 0:     # negative dimensions
        dim = data.ndim + dim
    size = data.shape[dim]  # R|I

    ddim = find_dic_dim(dic, dim)
    cplx = {'R': False, 'C': True}[dic['nptype'][ddim]]
    sw = dic['sw'][ddim]   # Hz
    obs = dic['sf'][ddim]   # MHz
    car = dic['ppm'][ddim] * dic[obs]   # Hz
    return fileiobase.unit_conversion(size, cplx, sw, obs, car)


#################
# data creation #
#################


def create_data(data):
    """
    Create a RNMRTK data array (recast into float32 or complex64)
    """
    if np.iscomplexobj(data):
        return np.array(data, dtype="complex64")
    else:
        return np.array(data, dtype="float32")

########################
# universal dictionary #
########################


def guess_udic(dic, data):
    """
    Guess parameters of a universal dictionary from a dic, data pair.

    Parameters
    ----------
    dic : dict
        Dictionary of RNMRTK parameters.
    data : ndarray
        Array of NMR data.

    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.

    """
    # create an empty universal dictionary
    ndim = dic['ndim']
    udic = fileiobase.create_blank_udic(ndim)

    # fill in parameters from RNMRTK dictionary for each dimension
    for iudim in range(ndim):

        # find the corresponding dimension in the RNMRTK parameter dictionary
        idim = find_dic_dim(dic, iudim)

        udic[iudim]['encoding'] = dic['quad'][idim].lower()
        udic[iudim]['car'] = dic['ppm'][idim] * dic['sf'][idim]
        udic[iudim]['obs'] = dic['sf'][idim]
        udic[iudim]['sw'] = dic['sw'][idim]
        udic[iudim]['size'] = dic['npts'][idim]

        # set quadrature and correct size
        if dic['nptype'][idim] == 'C':
            if iudim != ndim - 1:
                udic[iudim]['size'] *= 2    # don't double size of last dim
            udic[iudim]['complex'] = True
        else:
            udic[iudim]['complex'] = False

        # set label to T1 or F1, etc
        udic[iudim]['label'] = dic['dom'][idim] + str(idim + 1)

        # set time or frequency domain
        if dic['dom'][idim].upper() == 'T':
            udic[iudim]['freq'] = False
            udic[iudim]['time'] = True
        else:
            udic[iudim]['freq'] = True
            udic[iudim]['time'] = False

    return udic


def create_dic(udic, dim_order=None):
    """
    Create a RNMRTK dictionary from a universal dictionary.

    Parameters
    ----------
    udic : dict
        Universal dictionary of spectral parameters.
    dim_order : list, optional
        List mapping axis numbers in the universal dictionary to the order
        in which they will appear in the RNMRTK dictionary.  If None, the
        default [0, 1, 2, ...] is used.

    Returns
    --------
    dic : dict
        Dictionary of RNMRTK parameters.

    """

    # create the RNMRTK dictionary and fill with some default values
    dic = {}
    dic['comment'] = ''
    dic['ndim'] = ndim = int(udic['ndim'])
    dic['format'] = np.dtype('float32').str

    if dim_order is None:
        dim_order = range(ndim)  # default to 0, 1, 2, ...

    # set various parameters from the universal dictionary
    dic['dom'] = [['F', 'T'][udic[i]['time']] for i in dim_order]
    dic['nptype'] = [['R', 'C'][udic[i]['complex']] for i in dim_order]
    dic['ppm'] = [udic[i]['car'] / udic[i]['obs'] for i in dim_order]
    dic['sf'] = [udic[i]['obs'] for i in dim_order]
    dic['sw'] = [udic[i]['sw'] for i in dim_order]
    dic['npts'] = [udic[i]['size'] for i in dim_order]
    dic['quad'] = [udic[i]['encoding'].lower() for i in dim_order]

    # xfirst and xstep are freq domains values, correct later for time domain
    dic['xfirst'] = [-0.5 * i for i in dic['sw']]
    dic['xstep'] = [udic[i]['sw'] / udic[i]['size'] for i in dim_order]

    # these we guess on, they may be incorrect
    dic['cphase'] = [0.0] * ndim
    dic['lphase'] = [0.0] * ndim
    dic['nacq'] = [udic[i]['size'] for i in dim_order]

    # make small corrections as needed
    rnmrtk_quads = ['states', 'states-tppi', 'tppi', 'tppi-redfield']
    for i in range(ndim):

        # fix quadrature if not a valid RNMRTK quadrature
        if dic['quad'][i] not in rnmrtk_quads:
            dic['quad'][i] = 'states'

        # fix parameters if time domain data
        if dic['dom'][i] == 'T':    # time domain data
            dic['xfirst'][i] = 0.0
            if dic['quad'][i] in ['states']:
                # states time domain data
                dic['xstep'][i] = 1. / dic['sw'][i]
            else:
                # tppi time domain data
                dic['xstep'][i] = 0.5 / dic['sw'][i]

        # half the number of points if dimension is complex
        if dic['nptype'][i] == 'C':
            dic['npts'][i] //= 2

    # determine and set layout
    size = [udic[i]['size'] for i in range(ndim)]
    domains = [dic['dom'][i] + str(i + 1) for i in range(ndim)]

    # correct size of last dimension if complex
    if dic['nptype'][dim_order[-1]] == 'C':
        dic['npts'][dim_order[-1]] *= 2
        size[-1] *= 2
        if dic['dom'][dim_order[-1]] == 'F':
            dic['xstep'][dim_order[-1]] /= 2.

    dic['layout'] = (size, domains)

    return dic

#######################
# Reading and Writing #
#######################


def read(filename, par_file=None):
    """
    Read RNMRTK files.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file to read (.sec).
    par_file : str or None, optional
        Filename of RNMRTK parameter file. If None (default) a the last four
        characters of `file` are changed to .par.

    Returns
    -------
    dic : dic
        Dictionary of RNMRTK parameters.
    data : ndarray
        Array of NMR data.

    Notes
    -----
    The dictionary parameters are ordered opposite the data layout, that is to
    say the the FIRST parameter in each list corresponds to the LAST axis in
    the data array.

    See Also
    --------
    read_lowmem : Read RNMRTK files with minimal memory usage.
    write : Write RNMRTK files.

    """
    # determine par_file name if not given
    if par_file is None:
        par_file = filename[:-4] + ".par"
    dic = read_par(par_file)

    # determine sec file parameters from parameter dictionary
    dtype = dic["format"]
    shape = dic["layout"][0]
    cplex = {'R': False, 'C': True}[dic['nptype'][-1]]

    # read in the data
    data = read_sec(filename, dtype, shape, cplex)
    return dic, data


def read_lowmem(filename, par_file=None):
    """
    Read RNMRTK files with minimal memory usage

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file to read (.sec).
    par_file : str or None, optional
        Filename of RNMRTK parameter file. If None (default) a the last four
        characters of `file` are changed to .par.

    Returns
    -------
    dic : dic
        Dictionary of RNMRTK parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    Notes
    -----
    The dictionary parameters are ordered opposite the data layout, that is to
    say the the FIRST parameter in each list corresponds to the LAST axis in
    the data array.

    See Also
    --------
    read : Read RNMRTK files.
    write : Write RNMRTK files.

    """
    # determine par_file name if not given
    if par_file is None:
        par_file = filename[:-4] + ".par"
    dic = read_par(par_file)

    # determine shape, complexity and endiness from dictionary
    fshape = list(dic["layout"][0])
    cplex = {'R': False, 'C': True}[dic['nptype'][-1]]
    if cplex:
        fshape[-1] //= 2
    big = {'<': False, '>': True}[dic['format'][0]]
    data = rnmrtk_nd(filename, fshape, cplex, big)
    return dic, data


def write(filename, dic, data, par_file=None, overwrite=False):
    """
    Write RNMRTK files.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file to write to (.sec).
    dic : dict
        Dictionary of RNMRTK parameters.
    data : ndarray
        Array of NMR data.
    par_file : str or None, optional
        Filename of RNMRTK parameter file. If None (default) a the last four
        characters of `file` are changed to .par.
    overwrite : bool, optional
        True to overwrite existing files. False will raises a Warning if the
        file exists.

    See Also
    --------
    write_lowmem : Write RNMRTK files using minimal amounts of memory.
    read : Read RNMRTK files.

    """
    # determine par_file name if not given
    if par_file is None:
        par_file = filename[:-4] + ".par"
    write_par(par_file, dic, overwrite)
    dtype = dic["format"]
    write_sec(filename, data, dtype, overwrite)


def write_lowmem(filename, dic, data, par_file=None, overwrite=False):
    """
    Write RNMRTK files using minimal amounts of memory (trace by trace).

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file to write to (.sec).
    dic : dict
        Dictionary of RNMRTK parameters.
    data : array_like
        Array of NMR data.
    par_file : str or None, optional
        Filename of RNMRTK parameter file. If None (default) a the last four
        characters of `file` are changed to .par.
    overwrite : bool, optional
        True to overwrite existing files. False will raises a Warning if the
        file exists.

    See Also
    --------
    write : Write RNMRTK files using minimal amounts of memory.
    read_lowmem : Read RNMRTK files using minimal amounts of memory.

    """
    # determine par_file name if not given
    if par_file is None:
        par_file = filename[:-4] + ".par"
    write_par(par_file, dic, overwrite)

    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite, mode='wb')

    # write out the file trace by trace
    for tup in np.ndindex(data.shape[:-1]):
        put_trace(f, data[tup])
    f.close()
    return

#######################
# sec reading/writing #
#######################


def write_sec(filename, data, dtype='f4', overwrite=False):
    """
    Write a RNMRTK .sec file.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file to write to (.sec).
    data : array_like
        Array of NMR data.
    dtype : dtype
        Data type to convert data to before writing to disk.
    overwrite : bool, optional
        True to overwrite existing files. False will raises a Warning if the
        file exists.

    See Also
    --------
    write : Write RNMRTK files.

    """
    # open file
    f = fileiobase.open_towrite(filename, overwrite, mode='wb')

    # interleave read/imag if needed
    if np.iscomplexobj(data):
        data = interleave_data(data)

    # write data and close file
    f.write(data.astype(dtype).tostring())
    f.close()
    return


def read_sec(filename, dtype, shape, cplex):
    """
    Read a RNMRTK parameter .par file.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK (.sec) file to read .
    dtype : dtype
        Type of data in file, typically 'float32'.
    shape : tuple
        Shape of data.
    cplex : bool
        True if the last (fast) dimension is complex. False is real only.

    Returns
    -------
    data : ndarray
        Array of NMR data.

    """
    data = get_data(filename, dtype)
    data = data.reshape(shape)
    if cplex:
        data = uninterleave_data(data)
    return data

##########################
# data get/put functions #
##########################


def get_data(filename, dtype):
    """
    Get spectral data from a RNMRTK file.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file (.sec) to get data from.
    dtype : dtype
        Type of data in file, typically 'float32'

    Returns
    -------
    rdata : ndarray
        Raw NMR data, unshaped and typically not complex.

    """
    return np.fromfile(filename, dtype)


def get_trace(f, num_points, big):
    """
    Get a trace from an open RNMRTK file.

    Parameters
    -----------
    f : file object
        Open file object to read from.
    num_points : int
        Number of points in trace (R+I)
    big : bool
        True for data that is big-endian, False for little-endian.

    Returns
    -------
    trace : ndarray
        Raw trace of NMR data.

    """
    if big:
        bsize = num_points * np.dtype('>f4').itemsize
        return np.frombuffer(f.read(bsize), dtype='>f4')
    else:
        bsize = num_points * np.dtype('<f4').itemsize
        return np.frombuffer(f.read(bsize), dtype='<f4')


def put_trace(f, trace):
    """
    Put a trace to an open RNMRTK file.

    Parameters
    ----------
    f : file object
        Open file object to read from.
    trace : ndarray
        Raw trace of NMR data, may be complex64.

    """
    f.write(trace.view('float32').tostring())


def uninterleave_data(data):
    """
    Remove interleaving of real. imag data in last dimension of data.
    """
    return data.view('complex64')


def interleave_data(data):
    """
    Interleave real, imag data in data
    """
    return data.view('float32')
    # size = list(data.shape)
    # size[-1] = size[-1]*2
    # data_out = np.empty(size,dtype="float32")
    # data_out[...,::2] = data.real
    # data_out[...,1::2] = data.imag
    # return data_out

######################
# low-memory objects #
######################


class rnmrtk_nd(fileiobase.data_nd):
    """
    Emulate a ndarray objects without loading data into memory for low memory
    reading of RNMRTK files.

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes methods create a new objects with correct axes
      ordering.
    * has ndim, shape, and dtype attributes.

    Parameters
    ----------
    filename : str
        Filename of RNMRTK file (.sec) to read.
    fshape : tuple of ints
        Shape of data in file.
    cplex : bool
        True if the last (fast) axis is complex.
    big : bool
        True for big-endian data, False for little-endian.
    order : tuple
        Ordering of axes against file. None for (0, 1, 2, ...).

    """

    def __init__(self, filename, fshape, cplex, big, order=None):
        """
        Create and set up
        """
        # check and set order
        if order is None:
            order = range(len(fshape))
        self.order = order

        # set additional parameters
        self.fshape = fshape    # shape on disk
        self.cplex = cplex
        self.filename = filename
        self.big = big

        if self.cplex:
            self.dtype = np.dtype('complex64')
        else:
            self.dtype = np.dtype('float32')

        self.__setdimandshape__()   # set ndim and shape attributes

    def __fcopy__(self, order):
        """
        Create a copy
        """
        n = rnmrtk_nd(self.filename, self.filename, self.fshape, self.cplex,
                      self.big, order)
        return n

    def __fgetitem__(self, slices):
        """
        Return ndarray of selected values.

        slices is a well formatted tuple of slices

        """
        # seperate the last slice from the leading slices
        lslice = slices[-1]
        fslice = slices[:-1]

        # and the same for fshape
        lfshape = self.fshape[-1]
        ffshape = self.fshape[:-1]

        # find the output size and make an to/from nd_iterator
        osize, nd_iter = fileiobase.size_and_ndtofrom_iter(ffshape, fslice)
        osize.append(len(range(lfshape)[lslice]))

        # create an empty array to store the selected slices
        out = np.empty(tuple(osize), dtype=self.dtype)

        # opent the file for reading
        f = open(self.filename, 'rb')

        # read in the data trace by trace
        for out_index, in_index in nd_iter:

            # determine the trace number from the index
            ntrace = fileiobase.index2trace_flat(ffshape, in_index)

            # seek to the correct place in the file
            if self.cplex:
                ts = ntrace * lfshape * 2 * 4
                f.seek(ts)
                trace = get_trace(f, lfshape * 2, self.big)
                trace = uninterleave_data(trace)
            else:
                ts = ntrace * lfshape * 4
                f.seek(ts)
                trace = get_trace(f, lfshape, self.big)

            # put the trace into the output array
            out[out_index] = trace[lslice]

        # close the file and return
        f.close()
        return out

##################################
# Parameter dictionary utilities #
##################################


def find_dic_dim(dic, dim):
    """
    Find dimension in dictionary which corresponds to array dimension.

    Parameters
    ----------
    dic : dict
        Dictionary of RNMRTK parameters.
    dim : int, non-negative
        Dimension of data array.

    Returns
    -------
    ddim : int
        Dimension in dic which corresponds to array dimension, dim.

    """
    dic_dims = [int(i[1]) - 1 for i in dic['layout'][1]]
    return dic_dims.index(dim)


def find_array_dim(dic, ddim):
    """
    Find array dimension which corresponds to dictionary dimension.

    Parameters
    ----------
    dic : dict
        Dictionary of RNMRTK parameters.
    ddim : int, non-negative
        Dimension in dictionary.

    Returns
    -------
    dim : int
        Dimension in array which corresponds to dictionary dimension, ddim.

    """
    dic_dims = [int(i[1]) for i in dic['layout'][1]]
    return dic_dims[ddim]

############################
# parameter file functions #
############################


def read_par(filename):
    """
    Parse a RNMRTK parameter (.par) file.

    Parameters
    ----------
    file : str
        Filename of RNMRTK parameter file (.par) to read

    Returns
    -------
    dic : dict
        Dictionary of RNMRTK parameters.

    """
    dic = {}
    with open(filename, 'r') as f:
        for line in f:
            if len(line.split()) >= 2:
                parse_par_line(line, dic)

    # check that order and layout match, if they do remove from dictionary
    if dic['order'] != [int(i[1]) for i in dic['layout'][1]]:
        warn('Dom order and layout order do not match')
    else:
        dic.pop('order')

    return dic


def write_par(par_file, dic, overwrite):
    """
    Write a RNMRTK parameter file (.par).

    Parameters
    -----------
    par_file : str
        Filename of RNMRTK parameter file (.par) to write.
    dic : dict
        Dictionary of NMR parameters.
    overwrite : bool
        Set True to overwrite existing files, False will raise a Warning if the
        file exists.

    """
    # open file for writing
    f = fileiobase.open_towrite(par_file, overwrite, mode='w')

    # write comment line
    f.write('Comment \'' + dic['comment'] + '\'\n')

    # Dom line, set from layout
    l = "Dom " + " ".join(dic['layout'][1])
    f.write(l + "\n")

    # N line
    s = ["%14i %c" % (t) for t in zip(dic['npts'], dic['nptype'])]
    l = "N".ljust(8) + "".join(s)
    f.write(l + "\n")

    # write out additional lines Command    Value lines
    order = ['Sw', 'Xfirst', 'Xstep', 'Cphase', 'Lphase', 'Sf', 'Ppm', 'Nacq']
    codes = {'Sw': '%16.3f', 'Xfirst': '%16.5f', 'Xstep': '%16.5G',
             'Cphase': '%16.3f', 'Lphase': '%16.3f', 'Sf': '%16.2f',
             'Ppm': '%16.3f', 'Nacq': '%16i'}

    for lc in order:
        t = [codes[lc] % i for i in dic[lc.lower()]]
        l = lc.ljust(8) + "".join(t)
        f.write(l + "\n")

    # Quad line
    quad_dic = {'states': 'States', 'states-tppi': 'States-TPPI',
                'tppi': 'TPPI', 'tppi-redfield': 'TPPI-Redfield'}
    t = ["%16s" % (quad_dic[i]) for i in dic['quad']]
    l = "Quad".ljust(8) + "".join(t)
    f.write(l + "\n")

    # write format line
    if dic['format'][0] == '<':
        f.write('Format  Little-endian  IEEE-Float\n')
    else:
        f.write('Format  Big-endian     IEEE-Float\n')

    l = "Layout  " + " ".join([j + ":" + str(i) for i, j in
                               zip(*dic['layout'])])
    f.write(l + "\n")
    f.close()

    return


def parse_par_line(line, dic):
    """
    Parse a line from a RNMRTK parameter file (.par).
    """
    c, pl = line.split()[0], line.split()[1:]
    c = c.upper()

    if c == 'COMMENT':
        dic['comment'] = pl[0].strip('\'')

    elif c == 'DOM':
        dom = [s[0] for s in pl]        # dom as it appears in the file
        dic['ndim'] = ndim = len(pl)
        dic['order'] = order = [int(s[1]) for s in pl]  # dimension order
        # dom in accending order (to match other parameter)
        dic['dom'] = [dom[order.index(i)] for i in range(1, ndim + 1)]

    elif c == 'N':
        dic['npts'] = [int(i) for i in pl[::2]]
        dic['nptype'] = list(pl[1::2])

    elif c in ['SW', 'XFIRST', 'XSTEP', 'CPHASE', 'LPHASE', 'SF', 'PPM']:
        dic[c.lower()] = [float(i) for i in pl]
    elif c == "NACQ":
        dic['nacq'] = [int(i) for i in pl]
    elif c == "QUAD":
        dic['quad'] = [str(s).lower() for s in pl]

    # format assumes IEEE-Float type, only checks endiness
    elif c == 'FORMAT':
        if pl[0].upper() == "LITTLE-ENDIAN":
            dic['format'] = '<f4'
        else:
            dic['format'] = '>f4'
    elif c == 'LAYOUT':
        size = [int(p.split(":")[1]) for p in pl]
        domains = [p.split(":")[0] for p in pl]
        dic['layout'] = size, domains
    return



import numpy as np


def bruk_ranceY(dic, data, rotate_phase=True, **kwargs):
    """
    Reshuffle the data according to the Rance-Kay quadrature scheme.

    Parameters
    ----------
    dic : dict
        Dictionary of NMRPipe parameters
    data : ndarray
        Array of NMR data.
    rotate_phase : bool, optional
        Remove the requirement for a 90 degree zero-order phase correction

     Returns
    -------
    ndic : dict
        Dictionary of updated NMRPipe parameters.
    ndata : ndarray
        Array of NMR data which has been reshuffled according to the
        Rance-Kay scheme.

    """
    shuffled_data = np.empty(data.shape, data.dtype)

    for i in range(0, data.shape[0], 2):
        shuffled_data[i] = (1.*(data[i].real - data[i+1].real) +
                            1.*(data[i].imag - data[i+1].imag)*1j)
        if rotate_phase is True:
            shuffled_data[i+1] = (-1.*(data[i].imag + data[i+1].imag) +
                                  1.*(data[i].real + data[i+1].real)*1j)
        else:
            shuffled_data[i+1] = (1.*(data[i].real + data[i+1].real) +
                                  1.*(data[i].imag + data[i+1].imag)*1j)

    return dic, shuffled_data



"""
Functions for reading and writing files created by the `SIMPSON simulations
program <http://www.bionmr.chem.au.dk/bionmr/software/simpson.php>`_.
"""

import math
from warnings import warn

import numpy as np


def read(filename, ftype=None, ndim=None, NP=None, NI=None, spe=None):
    """
    Read a SIMPSON file.

    Read a NMR data file saved using in a number of formats produced using
    the SIMPSON simulation program.

    Parameters
    ----------
    filename : str
        Name of SIMPSON file to read data from.
    ftype : {None, 'TEXT', 'BINARY', 'XREIM', 'XYREIM', 'RAWBIN'}, optional
        A string indicating the type of SIMPSON file:

        ``TEXT`` : SIMPSON text format, no fsave arguments.

        ``BINARY`` : SIMPSON binary format, -binary fsave argument.

        ``XREIM`` : Indexed 1D format, rows of frequency/time, real and
        imaginary parts of the data.  Saved with -xreim argument.

        ``XYREIM`` : Indexed 2D format, rows of frequency/time (indirect
        dimension then direct dimension), real and imaginary parts of the
        data. Saved with -xyreim argument.

        ``RAWBIN`` : Raw binary format.  Saved with -raw_bin argument.  ndim
        and spe must also be provided.  In addition if ndim is 2, NP and NI
        must be defined.

        ``None`` : Automatically determine file type.  If this fails the file
        type should be implicitly provided.

        Other formats of files may be created by SIMPSON, but are not currently
        supported.

    Other Parameters
    ----------------
    ndim : {None, 1, 2}, optional
        Dimensionality of the data in the file, only used when ftype is
        "RAWBIN".
    NP : int, optional
        Number of points {R|I} in the direct dimension.  Only used when
        ftype is "RAWBIN" and ndim is 2.
    NI : int, optional
        Number of points in the indirect dimension.  Only used when ftype is
        "RAWBIN" and ndim is 2.
    spe : bool, optional
        True when the data is in the frequency domain, False for time domain
        data.  Only used when ftype is "RAWBIN"

    Returns
    -------
    dic : dict
        Dictionary of spectra parameters.  For some file formats this may be
        empty.
    data : ndarray
        Complex array of spectral data.

    """
    if ftype is None:
        ftype = guess_ftype(filename)

    if ftype == "TEXT":
        return read_text(filename)
    elif ftype == "BINARY":
        return read_binary(filename)
    elif ftype == "XREIM":
        return read_xreim(filename)
    elif ftype == "XYREIM":
        return read_xyreim(filename)
    elif ftype == "RAWBIN":
        if spe is None:
            raise ValueError('spe must be True or False for raw_bin data')
        if ndim == 1:
            return read_raw_bin_1d(filename, spe)
        elif ndim == 2:
            if NP is None or NI is None:
                raise ValueError("NP and NI must be given for raw_bin data")
            return read_raw_bin_2d(filename, NP, NI, spe)
        else:
            raise ValueError('ndim must be 1 or 2 for raw_bin data')
    else:
        raise ValueError("unknown ftype")


def guess_ftype(filename):
    """ Determine a SIMPSON file type from the first few lines of the file. """
    f = open(filename, 'rb')
    line = f.readline()

    if line[:4] == b"SIMP":   # either text or binary
        f.seek(0)
        for line in f:  # look for format line
            line = line.decode('ascii', 'ignore').strip('\n')
            if line[:6] == "FORMAT":
                f.close()
                return "BINARY"
            elif line == "DATA":    # early end if we screw up
                break
            else:
                continue
        f.close()
        return "TEXT"

    else:   # xyreim, xreim, or binary
        # determine the number of columns in the first 4 lines of the file
        ll1 = len(line.split())
        ll2 = len(f.readline().split())
        ll3 = len(f.readline().split())
        ll4 = len(f.readline().split())
        f.close()
        if ll1 == ll2 == ll3 == ll4 == 3:   # all have 3 columns
            return "XREIM"
        elif ll1 == ll2 == ll3 == ll4 == 4:  # all have 4 columns
            return "XYREIM"
        else:
            return "RAWBIN"

###############################
# Read SIMPSON text file data #
###############################


def read_text(filename):
    """Read a SIMPSON text file. See :py:func:`read`."""
    f = open(filename, 'r')  # open the file

    dic = {}                # initalize dictionary of parameters

    # parse the header of the file, storing parameters in dictionary, stop
    # when we hit the data line
    for line in f:
        line = line.strip('\n')
        if line == "SIMP":
            continue
        elif line == "DATA":
            break
        elif "=" in line:
            key, value = line.split("=")
            dic[key] = value
        else:
            warn("Warning, skipping line: %s" % (line))

    # convert float and int keys
    for key in ['SW1', 'SW']:    # convert keys to floats
        if key in dic:
            dic[key] = float(dic[key])
    for key in ['NP', 'NI', 'NELEM']:    # convert keys to ints
        if key in dic:
            dic[key] = int(dic[key])

    # set NELEM to 1 if its a single 1D/2D dataset
    if 'NELEM' not in dic:
        dic['NELEM'] = 1

    # DEBUGGING
    # return dic

    if "NI" in dic:     # 2D data
        data = np.empty((dic['NI']*dic['NELEM'], dic['NP']), dtype='complex64')

        # loop over remaining lines, extracting data values
        for iline, line in enumerate(f):
            if line[0:3] == "END":   # END marks end of data block
                break
            ni_idx, np_idx = divmod(iline, dic['NP'])
            r_val, i_val = [float(i) for i in line.split()]
            data.real[ni_idx, np_idx] = r_val
            data.imag[ni_idx, np_idx] = i_val

    else:   # 1D data
        data = np.empty((dic['NP']*dic['NELEM']), dtype='complex64')

        # loop over remaining lines, extracting data values
        for iline, line in enumerate(f):
            if line[0:3] == "END":
                break
            r_val, i_val = [float(i) for i in line.split()]
            data.real[iline] = r_val
            data.imag[iline] = i_val
        data = data.reshape(dic['NELEM'], -1)

    f.close()
    return dic, data


#############################
# Read xreim or xyreim data #
#############################


def read_xreim(filename):
    """Read a 1D indexed SIMPSON file.  See :py:func:`read`."""
    f = open(filename, 'r')
    # first pass, determine NP
    for iline, line in enumerate(f):
        pass
    NP = int(iline + 1)

    # second pass, place text-data into data and unit arrays
    f.seek(0)
    data = np.empty((NP, ), dtype='complex64')
    units = np.empty((NP, ), dtype='float32')

    for l_idx, line in enumerate(f):
        np_unit, r_val, i_val = [float(i) for i in line.split()]
        units[l_idx] = np_unit
        data.real[l_idx] = r_val
        data.imag[l_idx] = i_val

    f.close()
    return {'units': units}, data


def read_xyreim(filename):
    """Read a 2D indexed SIMPSON file. See :py:func:`read`."""
    f = open(filename, 'r')

    # first pass, determine NP, NI
    blines = 0  # number of blank lines, gives us num. of indirect points
    for nline, line in enumerate(f):
        if line == '\n':
            if blines == 0:     # location of first blank line gives us NP
                NP = int(nline)
            blines += 1
    NI = blines

    # second pass, place text-data into data and unit arrays
    f.seek(0)
    data = np.empty((NI, NP), dtype='complex64')
    units = np.recarray((NI, NP), dtype=[('ni_unit', 'f8'), ('np_unit', 'f8')])

    for l_idx, line in enumerate(f):
        ni_idx, np_idx = divmod(l_idx, NP + 1)  # determine indicies
        if np_idx == NP:    # skip blank line between blocks
            continue
        # unpack the line and store
        ni_unit, np_unit, r_val, i_val = [float(i) for i in line.split()]
        data.real[ni_idx, np_idx] = r_val
        data.imag[ni_idx, np_idx] = i_val
        units[ni_idx, np_idx].ni_unit = ni_unit
        units[ni_idx, np_idx].np_unit = np_unit

    f.close()
    return {'units': units}, data


######################
# Read raw_bin files #
######################


def read_raw_bin_1d(filename, spe=False):
    """Read a 1D raw binary SIMPSON file. See :py:func:`read`. """
    data = unappend_data(np.fromfile(filename, dtype='float32'))
    if spe:
        return {}, data[::-1]   # freq domain data is flipped
    else:
        return {}, data


def read_raw_bin_2d(filename, NP, NI, spe=False):
    """Read a 2D raw binary SIMPSON file. See :py:func:`read`. """
    data = np.fromfile(filename, dtype='float32').reshape(NI, NP * 2)
    if spe:
        # freq domain is saved reversed and shifted by 1 pt in each dimension,
        # so the first points is 0 in each vector, then N-1, N-2, ... 1
        # A bug? in SIMPSON causes the direct dimension shift to be left off
        # the imaginary data.  In addition the first real point in each vector
        # is the first real point in the spectrum.  Therefore only data[:,1:]
        # will match with other file formats.
        cdata = np.empty((NI, NP), dtype='complex64')
        cdata.real = np.roll(np.roll(data[::-1, NP - 1::-1], 1, 1), 1, 0)
        cdata.imag = np.roll(data[::-1, 2 * NP:NP - 1:-1], 1, 0)
        return {}, cdata
    else:
        # time domain is written as vectors of NP, with the imag appended
        return {}, unappend_data(data)


def unappend_data(data):
    """
    Return complex data with last axis unappended.

    Data should have imaginary data vector appended to real data vector.
    """
    h = int(data.shape[-1] / 2.0)
    return np.array(data[..., :h] + data[..., h:] * 1.j, dtype='complex64')


#############################
# Read SIMPSON binary files #
#############################


def read_binary(filename):
    """Read a binary SIMPSON file. See :py:func:`read`."""
    f = open(filename, 'r')  # open the file

    dic = {}                # initalize dictionary of parameters

    # parse the header of the file, storing parameters in dictionary, stop
    # when we hit the data line
    for line in f:
        line = line.strip('\n')
        if line == "SIMP":
            continue
        elif line == "DATA":
            break
        elif "=" in line:
            key, value = line.split("=")
            dic[key] = value
        else:
            warn("Warning, skipping line: %s" % (line))

    # convert float and int keys
    for key in ['SW1', 'SW']:    # convert keys to floats
        if key in dic:
            dic[key] = float(dic[key])
    for key in ['NP', 'NI', 'NELEM']:    # convert keys to ints
        if key in dic:
            dic[key] = int(dic[key])

    if not 'NELEM' in dic.keys():
        dic['NELEM'] = 1

    # DEBUGGING
    # return dic, f

    # extract characters from data block
    chardata = "".join([line.strip('\n') for line in f])
    chardata = chardata[:-3]    # remove END
    f.close()

    # convert every 4 character to 3 'bytes'
    nquads, mod = divmod(len(chardata), 4)
    assert mod == 0     # character should be in blocks of 4
    bytes = []
    for i in range(nquads):
        bytes += chars2bytes(chardata[i * 4:(i + 1) * 4])

    # DEBUGGING
    # return dic, bytes
    # convert every 4 'bytes' to a float, then complex data
    num_points, _num_pad = divmod(len(bytes), 4)
    data = np.empty((num_points, ), dtype='float32')

    for i in range(num_points):
        data[i] = bytes2float(bytes[i * 4: (i + 1) * 4])
    data = data.view('complex64')

    # reorder data according to dimensionality and domain
    if 'NI' in dic:  # 2D data
        return dic, data.reshape(dic['NI']*dic['NELEM'], dic['NP'])

    else:   # 1D data
        return dic, data.reshape(dic['NELEM'], dic['NP'])

BASE = 33
FIRST = lambda f, x: ((x) & ~(~0 << f))
LAST = lambda f, x: ((x) & (~0 << (8 - f)))


def chars2bytes(chars):
    """Convert four characters from a data block into 3 'bytes'."""
    c0, c1, c2, c3 = [ord(c) - BASE for c in chars]

    return [FIRST(6, c0) | LAST(2, c1 << 2),
            FIRST(4, c1) | LAST(4, c2 << 2),
            FIRST(2, c2) | LAST(6, c3 << 2)]


def bytes2float(bytes):
    """ Convert four bytes to a string. """
    # the first 23 bits (0-22) define the mantissa
    # the next 8 bits (23-30) the exponent,
    # the last bit (31) defines the sign
    b0, b1, b2, b3 = bytes
    mantissa = ((b2 % 128) << 16) + (b1 << 8) + b0
    exponent = (b3 % 128) * 2 + (b2 >= 128) * 1
    negative = b3 >= 128

    e = exponent - 0x7f
    m = np.abs(mantissa) / np.float64(1 << 23)

    if negative:
        return -math.ldexp(m, e)
    return math.ldexp(m, e)



"""
Misc. functions
"""

from __future__ import print_function

import numpy as np

# default tolerences
RTOL = 1.001e-01
ATOL = 1.001e-01
DTOL = 5.001e-01


def pair_similar(dic1, data1, dic2, data2, verb=False, atol=ATOL, rtol=RTOL,
                 dtol=DTOL, ignore_pipe_display=False):
    """
    Check a dic, data pair against a second dic, data pair for differences.

    Parameters
    ----------
    dic1 : dict
        First dictionary of NMR parameters.
    data1 : ndarray
        First array of NMR data
    dic2 : dict
        Second dictionary of NMR parameters
    data2 : ndarray
        Second array of NMR data
    verb : bool, optional
        Set True for verbose reporting.
    atol : float, optional
        The absolute tolerent parameter to pass to numpy.allclose.
    rtol : float, optional
        The relative tolenance parameter to pass to numpy.allclose.

    Returns
    -------
    r1 : bool
        True is data1 and data2 are similar, False if they differ.
    r2 : bool
        True is dic1 and dic2 are similar, False if they differ.

    """
    r1 = isdatasimilar(data1, data2, verb, atol, rtol)
    r2 = isdicsimilar(dict(dic1), dict(dic2), verb, dtol,
                      ignore_pipe_display=ignore_pipe_display)
    return r1, r2


def isdatasimilar(data1, data2, verb=False, atol=ATOL, rtol=RTOL):
    """
    Check that two sets of NMR data are equal within a tolerance.

    Parameters
    ----------
    data1 : ndarray
        First array of NMR data
    data2 : ndarray
        Second array of NMR data
    verb : bool, optional
        Set True for verbose reporting.
    atol : float, optional
        The absolute tolerent parameter to pass to numpy.allclose.
    rtol : float, optional
        The relative tolenance parameter to pass to numpy.allclose.

    Returns
    -------
    r1 : bool
        True is data1 and data2 are similar, False if they differ.

    """
    r = True
    if data1.dtype != data2.dtype:
        r = False
        if verb:
            print("Dtypes do not match:", data1.dtype, data2.dtype)
    if data1.shape != data2.shape:
        r = False
        if verb:
            print("Shapes do not match:", data1.shape, data2.shape)
    if np.allclose(data1, data2, rtol=rtol, atol=atol) is False:
        r = False
        if verb:
            print("Data does not match")
    return r


def isitemsimilar(v1, v2, verb=False, dtol=DTOL):
    """
    Compare two values for differences

    See :py:func:`isdicsimilar` for Parameters.

    """

    r = True

    # type checking
    if type(v1) != type(v2):
        r = False
        if verb:
            print("Item has different type", type(v1), type(v2))

    # iterable checking
    elif isinstance(v1, dict):
        r = r and isdicsimilar(v1, v2, verb=verb, dtol=dtol)

    elif isinstance(v1, list):
        r = r and islistsimilar(v1, v2, verb=verb, dtol=dtol)

    # numeric type
    elif isinstance(v1, (int, float)):
        if abs(v1 - v2) > dtol:
            r = False
            if verb:
                print("Key mismatch:", v1, v2)

    # all other types: just check if equal
    else:
        if v1 != v2:
            r = False
            if verb:
                print("Key mismatch:", v1, v2)

    return r


def isdicsimilar(dic1, dic2, verb=False, dtol=DTOL, ignore_pipe_display=False):
    """
    Compare two dictionaries for differences

    Float and int types compared within dtol. Lists and dictionaries are
    checked recursively all other checked by simple equivalence

    Parameters
    ----------
    dic1 : dict
        First dictionary of NMR parameters.
    dic2 : dict
        Second dictionary of NMR parameters
    verb : bool, optional
        Set True for verbose reporting.
    dtol : float, optional
        Maximum allowable difference between int and float elements if dic1
        and dic2.

    Returns
    -------
    r1 : bool
        True is dic1 and dic2 are similar, False if they differ.

    """
    # create copies of the two dictionaries
    dic1 = dict(dic1)
    dic2 = dict(dic2)

    # set return value to True
    r = True

    # create sets
    kset1 = set(dic1.keys())
    kset2 = set(dic2.keys())
    dset = set.difference(kset1, kset2)
    iset = set.intersection(kset1, kset2)

    if ignore_pipe_display is True:
        iset.discard('FDMIN')
        iset.discard('FDMAX')
        iset.discard('FDDISPMIN')
        iset.discard('FDDISPMAX')
        iset.discard('FDSCALEFLAG')

    # print out any keys not in both dictionaries
    if len(dset) != 0:
        r = False
        if verb:
            print("Keys not in both dictionaries:", dset)

    # loop over keys in both sets
    for k in iset:
        v1, v2 = dic1[k], dic2[k]

        if not isitemsimilar(v1, v2, verb=verb, dtol=dtol):
            print("For key:", k)
            r = False

    return r


def islistsimilar(l1, l2, verb=False, dtol=DTOL):
    """
    Compare two lists (or iterable) for differences

    See :py:func:`isdicsimilar` for Parameters.

    """
    # set return value to True
    r = True

    # print out any keys not in both dictionaries
    if len(l1) != len(l2):
        r = False
        if verb:
            print("Lists not of same length:", len(l1), len(l2))

    # loop over keys in both sets
    for v1, v2 in zip(l1, l2):
        if not isitemsimilar(v1, v2, verb=verb, dtol=dtol):
            r = False

    return r



from . import analysisbase
from . import leastsqbound
from . import segmentation
from . import peakpick
from . import linesh
from . import lineshapes1d
from . import helpers
from . import integration



"""
analysisbase provides general purpose analysis functions and classes used by
several nmrglue.analysis modules
"""

import numpy as np

pi = np.pi


# helper functions
def neighbors(pt, shape, structure):
    """
    Generate a list of all neighbors to a point.

    Parameters
    ----------
    pt : tuple of ints
        Index of the point to find neighbors of.
    shape : tuple of ints
        Shape of the region.
    structure : ndarray of bools
        Structure element that defines connections.

    Returns
    -------
    pts : list of int tuples
        List of tuples which represent indices for all points neighboring pt.
        Edges are treated as stopping points.

    """
    # set middle of structure to False
    s = np.copy(structure)  # copy structure
    middle = [int(np.floor(i / 2.)) for i in s.shape]  # find middle
    s.flat[np.ravel_multi_index(middle, s.shape)] = False
    offsets = np.argwhere(s) - middle

    # loop over the offset adding all valid points
    pts = []
    for offset in offsets:
        npt = pt - offset
        if valid_pt(npt, shape):
            pts.append(tuple(npt))
    return pts


def valid_pt(pt, shape):
    """
    Determind if a point (indices) is valid for a given shaped
    """
    for i, j in zip(pt, shape):
        if i < 0:   # index is not negative
            return False
        if i >= j:    # index is less than j
            return False
    return True

dimension_names = ['A', 'Z', 'Y', 'X']


# utility functions
def find_limits(pts):
    """
    Find the limits which outline the provided list of points

    Parameters
    ----------
    pts : list of int tuples
        List of points [(z0, y0, x0), (z1, y1, x1), ...]

    Returns
    -------
    min : ndarray
        Array of minimum indices: array([zmin, ymin, xmin]
    max : ndarray
        Array of maximum indices: array([zmin, ymin, xmin]

    See Also
    --------
    limits2slice : Create a list of slices from min, max limits

    """
    arr_pts = np.array(pts)
    return np.min(arr_pts, 0), np.max(arr_pts, 0)


def limits2slice(limits):
    """
    Create a set of slice objects given an array of min, max limits.

    Parameters
    ----------
    limits: tuple, (ndarray, ndarray)
        Two tuple consisting of array of the minimum and maximum indices.

    Returns
    -------
    slices : list
        List of slice objects which return points between limits

    See Also
    --------
    find_limits : Find the minimum and maximum limits from a list of points.
    slice2limits : Find a minimum and maximum limits for a list of slices.

    """
    mins, maxs = limits
    return tuple([slice(i, j + 1) for i, j in zip(mins, maxs)])


def slice2limits(slices):
    """
    Create a tuple of minimum, maximum limits from a set of slices.

    Parameters
    ----------
    slices : list
        List of slice objects which return points between limits

    Returns
    -------
    limits: tuple, (ndarray, ndarray)
        Two tuple consisting of array of the minimum and maximum indices.

    See Also
    --------
    limits2slice : Find a list of slices given minumum and maximum limits.
    """
    mins = [s.start for s in slices]
    maxs = [s.stop - 1 for s in slices]
    return mins, maxs


def squish(r, axis):
    """
    Squish array along an axis.

    Determine the sum along all but one axis for an array.

    Parameters
    ----------
    r : ndarray
        Array to squish.
    axis : int
        Axis of r to squish along.

    Returns
    -------
    s : 1D ndarray
        Array r squished into a single dimension.

    """
    # put axis to be squished as the last axis
    N = int(r.ndim)
    r = r.swapaxes(axis, N - 1)

    # sum along leading axis N-1 times
    for i in range(N - 1):
        r = r.sum(0)
    return r


# Windowing classes
class ndwindow(object):
    """
    An N-dimensional iterator to slice arrays into windows.

    Given the shape of an array and a window size, an 'ndwindow' instance
    iterators over tuples of slices which slice an the array into wsize
    sub-arrays.  At each iteration, the index of the center of the sub-array
    is incremented by one along the last dimension.  Array borders are ignored
    so the resulting sub-array can be smaller than wsize.  If wsize contains
    even values the window is off center containing an additional point with
    lower index.

    Parameters
    ----------
    size : tuple of ints
        Size of array to generate tuples of slices from.
    wsize : tuple of ints
        Window/sub-array size. Size of the area to select from array.  This is
        the maximum size of the window.

    Examples
    --------

    >>> a = np.arange(12).reshape(3,4)
    >>> for s in ndwindow(a.shape,(3,3)):
    ...     print(a[s])
    [[0 1]
     [4 5]]
    [[0 1 2]
     [4 5 6]]
    [[1 2 3]
     [5 6 7]]
    [[2 3]
     [6 7]]
    [[0 1]
     [4 5]
     [8 9]]
    [[ 0  1  2]
     [ 4  5  6]
     [ 8  9 10]]
    [[ 1  2  3]
     [ 5  6  7]
     [ 9 10 11]]
    [[ 2  3]
     [ 6  7]
     [10 11]]
    [[4 5]
     [8 9]]
    [[ 4  5  6]
     [ 8  9 10]]
    [[ 5  6  7]
     [ 9 10 11]]
    [[ 6  7]
     [10 11]]

    See Also
    --------
    ndwindow_index : Iterator of a ndwindow and index of the window center
    ndwindow_inside : Iterator over equal sized windows in the array.

    """
    def __init__(self, shape, wsize):
        """ Set up the ndwindow object """
        if len(shape) != len(wsize):
            raise ValueError("shape and wsize do match match")
        self.ndindex = np.ndindex(shape)
        wsize = np.array(wsize)
        self.sub = np.ceil((wsize - 1.) / 2.)
        self.add = wsize - 1. - self.sub

    def __next__(self):
        """ next iterator. """
        return self.next()

    def next(self):
        """ x.next() -> the next value, or raise StopIteration """
        center = self.ndindex.next()
        start = [max(0, i - j) for i, j in zip(center, self.sub)]
        stop = [i + j + 1 for i, j in zip(center, self.add)]
        return tuple([slice(x, y) for x, y in zip(start, stop)])

    def __iter__(self):
        """ x.__iter__() <==> iter(x) """
        return self


class ndwindow_index(object):
    """
    An N-dimensional interator object which returns the index of the window
    center and a :py:class:`ndwindow` slice array.  See :py:class:`ndwindow`
    for additional documentation.

    This class is equivalent to:

    for slices, index in zip(np.ndindex(shape), ndwindow(shape,wshape)):
        return (index, slice)

    See Also
    --------
    ndwindow: Iterator over only the window slices.
    ndwindow_inside : Iterator over equal sized windows in the array.

    """
    def __init__(self, shape, wsize):
        """ Set up the object """
        if len(shape) != len(wsize):
            raise ValueError("shape and wsize do match match")
        self.ndindex = np.ndindex(shape)
        wsize = np.array(wsize)
        self.sub = np.ceil((wsize - 1.) / 2.)
        self.add = wsize - 1. - self.sub

    def __next__(self):
        """ next iterator. """
        return self.next()

    def next(self):
        """ x.next() -> the next value, or raise StopIteration """
        center = self.ndindex.next()
        start = [max(0, i - j) for i, j in zip(center, self.sub)]
        stop = [i + j + 1 for i, j in zip(center, self.add)]
        return center, tuple([slice(x, y) for x, y in zip(start, stop)])

    def __iter__(self):
        """ x.__iter__() <==> iter(x) """
        return self


class ndwindow_inside(object):
    """
    An N-dimentional iterator to slice arrays into uniform size windows.

    Given the shape of an array and a window size, an 'ndwindow_inside'
    instance iterators over tuples of slices which slice an the array into
    uniform size wsize windows/sub-arrays.  At each iteration, the index of
    the top left of the sub-array is incremented by one along the last
    dimension utill the resulting windows would extend past the array border.
    All sub-arrays are equal sized (wsize).

    Parameters
    ----------
    size : tuple of ints
        Size of array to generate tuples of slices from.
    wsize : tuple of ints
        Size of the area to select from array (widow size).

    Examples
    --------

    >>> a = np.arange(9).reshape(3,3)
    >>> for s in ndwindow_inside(a.shape,(2,2)):
    ...     print(a[s])
    [[0 1]
     [3 4]]
    [[1 2]
     [4 5]]
    [[3 4]
     [6 7]]
    [[4 5]
     [7 8]]

    See Also
    --------
    ndwindow : Iterator over non-uniform windows.
    ndwindow_inside_index : Iterator of a ndwindow_inside and the index of the
        window's top left point.

    """
    def __init__(self, shape, wsize):
        """ Set up the object """
        if len(shape) != len(wsize):
            raise ValueError("shape and wsize do match match")
        self.ndindex = np.ndindex(
            tuple(np.array(shape) - np.array(wsize) + 1))
        self.wsize = wsize

    def __next__(self):
        """ next iterator. """
        return self.next()

    def next(self):
        """ x.next() -> the next value, or raise StopIteration """
        start = self.ndindex.next()
        stop = np.array(start) + np.array(self.wsize)
        return tuple([slice(x, y) for x, y in zip(start, stop)])

    def __iter__(self):
        """ x.__iter__() <==> iter(x) """
        return self


class ndwindow_inside_index(object):
    """
    An N-dimensional interator object which returns the index of the window
    top-left and a :py:class:`ndwindow_inside` slice array.

    Similar to :py:class:`ndwindow_index` but reports top left index of
    window.

    See :py:class:`ndwindow_inside` and :py:class`ndwindow_index` for addition
    documentation.

    """
    def __init__(self, shape, wsize):
        " Set up the object """
        if len(shape) != len(wsize):
            raise ValueError("shape and wsize do match match")
        self.ndindex = np.ndindex(
            tuple(np.array(shape) - np.array(wsize) + 1))
        self.wsize = wsize

    def __next__(self):
        """ next iterator. """
        return self.next()

    def next(self):
        """ x.next() -> the next value, or raiseStopIteration """
        start = self.ndindex.next()
        stop = np.array(start) + np.array(self.wsize)
        return (start, tuple([slice(x, y) for x, y in zip(start, stop)]))

    def __iter__(self):
        """ x.__iter__() <==> iter(x) """
        return self
